import lhcb_rex.settings.globals as myGlobals
import numpy as np
import lhcb_rex.tools.variables_tools as vt
import uproot
import lhcb_rex.tools.tuple_tools as tt
import vector
import pickle

# pid_list = [11, 13, 211, 321]
pid_list = [11, 13, 211, 321, 2212]
PID_charges = {
    11: -1,
    -11: 1,
    13: -1,
    -13: 1,
    211: 1,
    -211: -1,
    321: 1,
    -321: -1,
    2212: 1,
    -2212: -1,
}


def compute_masses(data_in, units="MeV"):
    particles = ["DAUGHTER1", "DAUGHTER2", "DAUGHTER3"]

    masses = {}
    if units == "MeV":
        masses[2212] = 938.272
        masses[321] = 493.677
        masses[211] = 139.57039
        masses[13] = 105.66
        masses[11] = 0.51099895000  # * 1e-3
    elif units == "GeV":
        masses[2212] = 938.272 * 1e-3
        masses[321] = 493.677 * 1e-3
        masses[211] = 139.57039 * 1e-3
        masses[13] = 105.66 * 1e-3
        masses[11] = 0.51099895000 * 1e-3

    for particle in particles:
        mass = np.asarray(data_in[f"{particle}_ID"]).astype("float32")
        for pid in pid_list:
            mass[np.where(np.abs(mass) == pid)] = masses[pid]
        data_in[f"{particle}_mass"] = mass

    for particle in particles:
        charge_ = np.asarray(data_in[f"{particle}_ID"]).copy()
        for key in list(PID_charges.keys()):
            charge_[np.where(charge_ == key)] = PID_charges[key]
        data_in[f"{particle}_charge"] = charge_

    where_keep = np.where(
        data_in[f"{particles[0]}_charge"] != data_in[f"{particles[1]}_charge"]
    )
    where_swap = np.where(
        data_in[f"{particles[0]}_charge"] == data_in[f"{particles[1]}_charge"]
    )

    ########################
    # Retrieve momenta (RECONSTRUCTED) and masses
    px1_reco, py1_reco, pz1_reco = (
        np.asarray(data_in[f"{particles[0]}_PX"]).copy(),
        np.asarray(data_in[f"{particles[0]}_PY"]).copy(),
        np.asarray(data_in[f"{particles[0]}_PZ"]).copy(),
    )
    px2_reco, py2_reco, pz2_reco = (
        np.asarray(data_in[f"{particles[1]}_PX"]).copy(),
        np.asarray(data_in[f"{particles[1]}_PY"]).copy(),
        np.asarray(data_in[f"{particles[1]}_PZ"]).copy(),
    )
    px3_reco, py3_reco, pz3_reco = (
        np.asarray(data_in[f"{particles[2]}_PX"]).copy(),
        np.asarray(data_in[f"{particles[2]}_PY"]).copy(),
        np.asarray(data_in[f"{particles[2]}_PZ"]).copy(),
    )

    px2_reco_buffer, py2_reco_buffer, pz2_reco_buffer = (
        px2_reco.copy(),
        py2_reco.copy(),
        pz2_reco.copy(),
    )
    px3_reco_buffer, py3_reco_buffer, pz3_reco_buffer = (
        px3_reco.copy(),
        py3_reco.copy(),
        pz3_reco.copy(),
    )

    px2_reco[where_swap], py2_reco[where_swap], pz2_reco[where_swap] = (
        px3_reco_buffer[where_swap],
        py3_reco_buffer[where_swap],
        pz3_reco_buffer[where_swap],
    )
    px3_reco[where_swap], py3_reco[where_swap], pz3_reco[where_swap] = (
        px2_reco_buffer[where_swap],
        py2_reco_buffer[where_swap],
        pz2_reco_buffer[where_swap],
    )
    ########################

    ########################
    # Retrieve momenta (TRUE) and masses
    try:
        px1, py1, pz1 = (
            np.asarray(data_in[f"{particles[0]}_TRUEP_X"]).copy(),
            np.asarray(data_in[f"{particles[0]}_TRUEP_Y"]).copy(),
            np.asarray(data_in[f"{particles[0]}_TRUEP_Z"]).copy(),
        )
        px2, py2, pz2 = (
            np.asarray(data_in[f"{particles[1]}_TRUEP_X"]).copy(),
            np.asarray(data_in[f"{particles[1]}_TRUEP_Y"]).copy(),
            np.asarray(data_in[f"{particles[1]}_TRUEP_Z"]).copy(),
        )
        px3, py3, pz3 = (
            np.asarray(data_in[f"{particles[2]}_TRUEP_X"]).copy(),
            np.asarray(data_in[f"{particles[2]}_TRUEP_Y"]).copy(),
            np.asarray(data_in[f"{particles[2]}_TRUEP_Z"]).copy(),
        )
    except:
        px1, py1, pz1 = (
            np.asarray(data_in[f"{particles[0]}_PX_TRUE"]).copy(),
            np.asarray(data_in[f"{particles[0]}_PY_TRUE"]).copy(),
            np.asarray(data_in[f"{particles[0]}_PZ_TRUE"]).copy(),
        )
        px2, py2, pz2 = (
            np.asarray(data_in[f"{particles[1]}_PX_TRUE"]).copy(),
            np.asarray(data_in[f"{particles[1]}_PY_TRUE"]).copy(),
            np.asarray(data_in[f"{particles[1]}_PZ_TRUE"]).copy(),
        )
        px3, py3, pz3 = (
            np.asarray(data_in[f"{particles[2]}_PX_TRUE"]).copy(),
            np.asarray(data_in[f"{particles[2]}_PY_TRUE"]).copy(),
            np.asarray(data_in[f"{particles[2]}_PZ_TRUE"]).copy(),
        )

    px2_buffer, py2_buffer, pz2_buffer = px2.copy(), py2.copy(), pz2.copy()
    px3_buffer, py3_buffer, pz3_buffer = px3.copy(), py3.copy(), pz3.copy()

    px2[where_swap], py2[where_swap], pz2[where_swap] = (
        px3_buffer[where_swap],
        py3_buffer[where_swap],
        pz3_buffer[where_swap],
    )
    px3[where_swap], py3[where_swap], pz3[where_swap] = (
        px2_buffer[where_swap],
        py2_buffer[where_swap],
        pz2_buffer[where_swap],
    )
    ########################

    mass_shape = np.shape(np.asarray(data_in[f"{particles[0]}_mass"]).copy())
    mass1, mass2, mass3 = (
        np.ones(mass_shape) * masses[321],
        np.ones(mass_shape) * masses[11],
        np.ones(mass_shape) * masses[11],
    )

    mass2_buffer, mass3_buffer = mass2.copy(), mass3.copy()
    mass2_buffer[where_swap], mass3_buffer[where_swap] = (
        mass3_buffer[where_swap],
        mass2_buffer[where_swap],
    )

    # Calculate energies
    E1 = np.sqrt(px1**2 + py1**2 + pz1**2 + mass1**2)
    E2 = np.sqrt(px2**2 + py2**2 + pz2**2 + mass2**2)
    E3 = np.sqrt(px3**2 + py3**2 + pz3**2 + mass3**2)

    # Calculate energies
    E1_reco = np.sqrt(px1_reco**2 + py1_reco**2 + pz1_reco**2 + mass1**2)
    E2_reco = np.sqrt(px2_reco**2 + py2_reco**2 + pz2_reco**2 + mass2**2)
    E3_reco = np.sqrt(px3_reco**2 + py3_reco**2 + pz3_reco**2 + mass3**2)

    # Compute invariant mass squared for each pair
    def invariant_mass_squared(E1, E2, px1, px2, py1, py2, pz1, pz2):
        return (E1 + E2) ** 2 - ((px1 + px2) ** 2 + (py1 + py2) ** 2 + (pz1 + pz2) ** 2)

    # Compute invariant mass squared for each pair
    def invariant_mass_squared_3_particles(
        E1, E2, E3, px1, px2, px3, py1, py2, py3, pz1, pz2, pz3
    ):
        return (E1 + E2 + E3) ** 2 - (
            (px1 + px2 + px3) ** 2 + (py1 + py2 + py3) ** 2 + (pz1 + pz2 + pz3) ** 2
        )

    def compute_costhetal_vec(E1, E2, E3, px1, px2, px3, py1, py2, py3, pz1, pz2, pz3):
        B = vector.obj(
            px=px1 + px2 + px3, py=py1 + py2 + py3, pz=pz1 + pz2 + pz3, E=E1 + E2 + E3
        )
        eplus = vector.obj(px=px3, py=py3, pz=pz3, E=E3)
        eminus = vector.obj(px=px2, py=py2, pz=pz2, E=E2)

        vecA = eplus.boost_beta3(-B.to_beta3())
        vecB = eminus.boost_beta3(-B.to_beta3())
        vecAB = vecA.add(vecB)

        boostedA = vecB.boost_beta3(-vecAB.to_beta3())
        vecAB.E = 0
        boostedA.E = 0

        dirAB = vecAB.unit()
        dirboostedA = boostedA.unit()
        costhetal = dirboostedA.dot(dirAB)

        return costhetal

    # m12^2, m13^2, and m23^2
    m12_squared = invariant_mass_squared(E1, E2, px1, px2, py1, py2, pz1, pz2)
    m13_squared = invariant_mass_squared(E1, E3, px1, px3, py1, py3, pz1, pz3)
    m23_squared = invariant_mass_squared(E2, E3, px2, px3, py2, py3, pz2, pz3)

    m123_squared = invariant_mass_squared_3_particles(
        E1, E2, E3, px1, px2, px3, py1, py2, py3, pz1, pz2, pz3
    )

    ctl = compute_costhetal_vec(E1, E2, E3, px1, px2, px3, py1, py2, py3, pz1, pz2, pz3)

    # m12^2, m13^2, and m23^2
    m12_squared_reco = invariant_mass_squared(
        E1_reco, E2_reco, px1_reco, px2_reco, py1_reco, py2_reco, pz1_reco, pz2_reco
    )
    m13_squared_reco = invariant_mass_squared(
        E1_reco, E3_reco, px1_reco, px3_reco, py1_reco, py3_reco, pz1_reco, pz3_reco
    )
    m23_squared_reco = invariant_mass_squared(
        E2_reco, E3_reco, px2_reco, px3_reco, py2_reco, py3_reco, pz2_reco, pz3_reco
    )

    m123_squared_reco = invariant_mass_squared_3_particles(
        E1_reco,
        E2_reco,
        E3_reco,
        px1_reco,
        px2_reco,
        px3_reco,
        py1_reco,
        py2_reco,
        py3_reco,
        pz1_reco,
        pz2_reco,
        pz3_reco,
    )

    ctl_reco = compute_costhetal_vec(
        E1_reco,
        E2_reco,
        E3_reco,
        px1_reco,
        px2_reco,
        px3_reco,
        py1_reco,
        py2_reco,
        py3_reco,
        pz1_reco,
        pz2_reco,
        pz3_reco,
    )

    # print(np.sqrt(m123_squared))
    # print(np.sqrt(m123_squared_reco))
    # quit()

    data_in["dalitz_mass_m12"] = np.asarray(m12_squared) / 1e6
    data_in["dalitz_mass_m13"] = np.asarray(m13_squared) / 1e6
    data_in["dalitz_mass_m23"] = np.asarray(m23_squared) / 1e6

    data_in["dalitz_mass_m12_reco"] = np.asarray(m12_squared_reco) / 1e6
    data_in["dalitz_mass_m13_reco"] = np.asarray(m13_squared_reco) / 1e6
    data_in["dalitz_mass_m23_reco"] = np.asarray(m23_squared_reco) / 1e6

    data_in["mkl"] = np.sqrt(np.asarray(m12_squared_reco))
    data_in["q2"] = np.asarray(m23_squared_reco) / 1e6
    data_in["mkee"] = np.sqrt(np.asarray(m123_squared_reco)) / 1e3

    data_in["ctl_true"] = ctl
    data_in["ctl"] = ctl_reco

    return data_in


def add_conditions(events, particles, mother, intermediates, intermediate_recipes):
    # # Kee physics example validation variables:
    # events = compute_masses(events)

    for particle in particles:
        # add true mass values for each particle based on TRUEID
        mass = np.asarray(events[f"{particle}_TRUEID"]).astype("float32")
        for pid in pid_list:
            mass[np.where(np.abs(mass) == pid)] = myGlobals.masses[pid]
        events[f"{particle}_mass"] = mass

    # convert ENDVERTEX_CHI2 to ENDVERTEX_CHI2NDOF
    events[f"{mother}_ENDVERTEX_CHI2NDOF"] = (
        events[f"{mother}_ENDVERTEX_CHI2"] / events[f"{mother}_ENDVERTEX_NDOF"]
    )
    for intermediate in intermediates:
        events[f"{intermediate}_ENDVERTEX_CHI2NDOF"] = (
            events[f"{intermediate}_ENDVERTEX_CHI2"]
            / events[f"{intermediate}_ENDVERTEX_NDOF"]
        )

    events[f"{mother}_TRUE_FD"] = events.eval(
        "sqrt((MOTHER_TRUEENDVERTEX_X-MOTHER_TRUEORIGINVERTEX_X)**2 + (MOTHER_TRUEENDVERTEX_Y-MOTHER_TRUEORIGINVERTEX_Y)**2 + (MOTHER_TRUEENDVERTEX_Z-MOTHER_TRUEORIGINVERTEX_Z)**2)"
    )

    # compute distance mother flies before decay
    A = vt.compute_distance(events, mother, "TRUEENDVERTEX", mother, "TRUEORIGINVERTEX")
    A = np.asarray(A)
    A[np.where(A == 0)] = 5e-5
    events[f"{mother}_FLIGHT"] = A

    # compute distance (scalar) any intermediate flies before decay
    for particle in particles:
        A = vt.compute_distance(
            events, particle, "TRUEORIGINVERTEX", mother, "TRUEENDVERTEX"
        )
        A = np.asarray(A)
        A[np.where(A == 0)] = 5e-5
        events[f"{particle}_FLIGHT"] = A

    # print(events)
    # quit()
    # for particle in particles:
    #     for dim in ["X",'Y',"Z"]:
    #         events[f"{particle}_TRUEORIGIN_wrtMEV_VERTEX_X"] = events[f"{particle}_TRUEORIGINVERTEX_X"]-events[f"{mother}_TRUEENDVERTEX_X"]
    #     events[f"{particle}_TRUEORIGIN_wrtMEV_VERTEX_T"] =

    # compute distance any intermediate flies before decay in each dimension
    for particle in particles:
        events[f"{particle}_deltaEV_X"] = (
            events[f"{particle}_TRUEORIGINVERTEX_X"]
            - events[f"{mother}_TRUEENDVERTEX_X"]
        )
        events[f"{particle}_deltaEV_Y"] = (
            events[f"{particle}_TRUEORIGINVERTEX_Y"]
            - events[f"{mother}_TRUEENDVERTEX_Y"]
        )
        events[f"{particle}_deltaEV_Z"] = (
            events[f"{particle}_TRUEORIGINVERTEX_Z"]
            - events[f"{mother}_TRUEENDVERTEX_Z"]
        )
        events[f"{particle}_deltaEV_T"] = np.sqrt(
            events[f"{particle}_deltaEV_X"] ** 2 + events[f"{particle}_deltaEV_Y"] ** 2
        )

    # compute masses of mother particle, reco and true
    events[f"{mother}_M"] = vt.compute_mass_N(events, particles)
    events[f"{mother}_M_reco"] = vt.compute_mass_N(events, particles, true_vars=False)

    # compute whether fully reco or not
    mother_masses = -1.0 * np.ones(events.shape[0])
    mother_TRUEID = events["MOTHER_TRUEID"]
    for PID in list(myGlobals.particle_masses_dict_mothers.keys()):
        mother_masses[np.where(np.abs(mother_TRUEID.astype(int)) == PID)] = (
            myGlobals.particle_masses_dict_mothers[PID]
        )
    events["abs_mass_diff"] = np.abs(np.asarray(events["MOTHER_M"]) - mother_masses)
    fully_reco = np.zeros(events.shape[0])
    fully_reco[np.where(events["abs_mass_diff"] < 50.0)] = 1  # 1 if within 50 MeV
    events["fully_reco"] = fully_reco

    # comptue mother momenta, reco and true
    events[f"{mother}_P"], events[f"{mother}_PT"] = (
        vt.compute_reconstructed_mother_momenta(events, mother, true_vars=False)
    )
    events[f"{mother}_TRUEP"], events[f"{mother}_TRUEP_T"] = (
        vt.compute_reconstructed_mother_momenta(events, mother, true_vars=True)
    )

    # comptue residual in momentum measurement of each particle
    for particle in particles:
        for dim in ["X", "Y", "Z"]:
            residual_frac = (
                events[f"{particle}_P{dim}"] - events[f"{particle}_TRUEP_{dim}"]
            ) / (events[f"{particle}_TRUEP_{dim}"] + 1e-4)

            _residualfrac_limit = 5.0
            limit = _residualfrac_limit
            residual_frac[residual_frac < (limit * -1.0)] = -limit
            residual_frac[residual_frac > limit] = limit

            events[f"{particle}_residualfrac_P{dim}"] = residual_frac

        PX = events[f"{particle}_TRUEP_X"]
        PY = events[f"{particle}_TRUEP_Y"]
        PZ = events[f"{particle}_TRUEP_Z"]
        PX_reco = events[f"{particle}_PX"]
        PY_reco = events[f"{particle}_PY"]
        PZ_reco = events[f"{particle}_PZ"]

        events[f"{particle}_P"] = np.sqrt((PX_reco**2 + PY_reco**2 + PZ_reco**2))
        events[f"{particle}_TRUEP"] = np.sqrt((PX**2 + PY**2 + PZ**2))
        events[f"{particle}_Preco_overP"] = (
            events[f"{particle}_P"] / events[f"{particle}_TRUEP"]
        )

    # comptue how much momentum is missing from the reconstructed mother
    events[f"{mother}_missing_P"], events[f"{mother}_missing_PT"] = (
        vt.compute_missing_momentum(events, mother, particles)
    )
    for intermediate in intermediates:
        events[f"{intermediate}_P"], events[f"{intermediate}_PT"] = (
            vt.compute_reconstructed_intermediate_momenta(
                events, intermediate_recipes[intermediate], true_vars=False
            )
        )

        events[f"{intermediate}_TRUEP"], events[f"{intermediate}_TRUEP_T"] = (
            vt.compute_reconstructed_intermediate_momenta(
                events, intermediate_recipes[intermediate], true_vars=True
            )
        )

        events[f"{intermediate}_TRUEP_X"] = np.zeros(
            np.shape(events["DAUGHTER2_TRUEP_X"])
        )
        events[f"{intermediate}_TRUEP_Y"] = np.zeros(
            np.shape(events["DAUGHTER2_TRUEP_Y"])
        )
        events[f"{intermediate}_TRUEP_Z"] = np.zeros(
            np.shape(events["DAUGHTER2_TRUEP_Z"])
        )
        for particle in intermediate_recipes[intermediate]:
            events[f"{intermediate}_TRUEP_X"] += events[f"{particle}_TRUEP_X"]
            events[f"{intermediate}_TRUEP_Y"] += events[f"{particle}_TRUEP_Y"]
            events[f"{intermediate}_TRUEP_Z"] += events[f"{particle}_TRUEP_Z"]

        events[f"{intermediate}_PX"] = np.zeros(np.shape(events["DAUGHTER2_PX"]))
        events[f"{intermediate}_PY"] = np.zeros(np.shape(events["DAUGHTER2_PY"]))
        events[f"{intermediate}_PZ"] = np.zeros(np.shape(events["DAUGHTER2_PZ"]))
        for particle in intermediate_recipes[intermediate]:
            events[f"{intermediate}_PX"] += events[f"{particle}_PX"]
            events[f"{intermediate}_PY"] += events[f"{particle}_PY"]
            events[f"{intermediate}_PZ"] += events[f"{particle}_PZ"]

    for particle in particles:
        events[f"{particle}_TRUEE"] = vt.compute_TRUEE(events, particle)

        (
            events[f"{particle}_delta_P"],
            events[f"{particle}_delta_PT"],
        ) = vt.compute_reconstructed_momentum_residual(events, particle)

    # print(events)
    # quit()

    for particle in particles:
        # if f"{particle}_TRUEP_T" not in events:
        PT = events.eval(f"sqrt({particle}_TRUEP_X**2 + {particle}_TRUEP_Y**2)")
        # else:
        #     PT = events[f"{particle}_TRUEP_T"]
        # if f"{particle}_TRUEP" not in events:
        P = events.eval(
            f"sqrt({particle}_TRUEP_X**2 + {particle}_TRUEP_Y**2 + {particle}_TRUEP_Z**2)"
        )
        # else:
        #     P = events[f"{particle}_TRUEP"]

        events[f"{particle}_eta_TRUE"] = -np.log(np.tan(np.arcsin(PT / P) / 2.0))

    for particle in particles:
        # if f"{particle}_PT" not in events:
        PT = events.eval(f"sqrt({particle}_PX**2 + {particle}_PY**2)")
        # else:
        #     PT = events[f"{particle}_PT"]
        # if f"{particle}_P" not in events:
        P = events.eval(f"sqrt({particle}_PX**2 + {particle}_PY**2 + {particle}_PZ**2)")
        # else:
        #     P = events[f"{particle}_P"]

        events[f"{particle}_eta"] = -np.log(np.tan(np.arcsin(PT / P) / 2.0))

    for particle in particles:
        for dim in ["X", "Y", "Z"]:
            events[f"{particle}_delta_P{dim}"] = (
                events[f"{particle}_P{dim}"] - events[f"{particle}_TRUEP_{dim}"]
            )
        events[f"{particle}_delta_eta"] = (
            events[f"{particle}_eta"] - events[f"{particle}_eta_TRUE"]
        )
        # events[f"{particle}_delta_theta"] = np.arctan2(events[f"{particle}_delta_PX"], events[f"{particle}_delta_PY"])

    ################################################################################

    for particle in particles:
        events[f"{particle}_angle_wrt_mother"] = vt.compute_angle(
            events, mother, f"{particle}"
        )
        events[f"{particle}_angle_wrt_mother_reco"] = vt.compute_angle(
            events, mother, f"{particle}", true_vars=False
        )
        # for intermediate in intermediates:
        #     events[f"{particle}_angle_wrt_mother_intermediate"] = vt.compute_angle(
        #         events, intermediate, f"{particle}"
        #     )
        #     events[f"{particle}_angle_wrt_mother_intermediate_reco"] = vt.compute_angle(
        #         events, intermediate, f"{particle}", true_vars=False
        #     )

    for i, particle_i in enumerate(particles):
        for j, particle_j in enumerate(particles):
            if i != j:
                events[f"edge_angle_{particle_i}_{particle_j}"] = vt.compute_angle(
                    events, f"{particle_i}", f"{particle_j}", true_vars=False
                )
                events[f"edge_angle_{particle_i}_{particle_j}_TRUE"] = vt.compute_angle(
                    events, f"{particle_i}", f"{particle_j}", true_vars=True
                )

    true_vertex = True
    events[f"{mother}_IP"] = vt.compute_impactParameter(
        events, mother, particles, true_vertex=true_vertex, true_vars=False
    )
    for particle in particles:
        events[f"{particle}_IP"] = vt.compute_impactParameter_i(
            events, mother, f"{particle}", true_vertex=true_vertex, true_vars=False
        )
    events[f"{mother}_DIRA"] = vt.compute_DIRA(
        events, mother, particles, true_vertex=true_vertex, true_vars=False
    )

    events[f"{mother}_IP_TRUE"] = vt.compute_impactParameter(
        events, mother, particles, true_vertex=true_vertex, true_vars=True
    )
    for particle in particles:
        events[f"{particle}_IP_TRUE"] = vt.compute_impactParameter_i(
            events, mother, f"{particle}", true_vertex=true_vertex, true_vars=True
        )
    events[f"{mother}_DIRA_TRUE"] = vt.compute_DIRA(
        events, mother, particles, true_vertex=true_vertex, true_vars=True
    )

    for intermediate in intermediates:
        events[f"{intermediate}_IP"] = vt.compute_impactParameter(
            events,
            mother,
            intermediate_recipes[intermediate],
            true_vertex=true_vertex,
            true_vars=False,
        )
        events[f"{intermediate}_DIRA"] = vt.compute_DIRA(
            events,
            mother,
            intermediate_recipes[intermediate],
            true_vertex=true_vertex,
            true_vars=False,
        )

        events[f"{intermediate}_IP_TRUE"] = vt.compute_impactParameter(
            events,
            mother,
            intermediate_recipes[intermediate],
            true_vertex=true_vertex,
            true_vars=True,
        )
        events[f"{intermediate}_DIRA_TRUE"] = vt.compute_DIRA(
            events,
            mother,
            intermediate_recipes[intermediate],
            true_vertex=true_vertex,
            true_vars=True,
        )

    events["weight"] = 1.0
    weights = {}
    weights[11] = 11.0
    weights[2212] = 1.0
    weights[211] = 0.15
    weights[321] = 0.85
    weights[13] = 1.7
    for ID in [11, 211, 321, 13, 2212]:
        sel_string = ""
        for particle in particles:
            sel_string += f"(abs({particle}_TRUEID) == {ID})"
            sel_string += " or "
        sel_string = sel_string[:-4]
        select = events.query(sel_string)
        events.loc[select.index, "weight"] *= weights[ID]

    return events


def compute_and_append(file_name, setting_file):
    head = -1

    print(setting_file)

    with open(setting_file, "rb") as handle:
        setting = pickle.load(handle)
        setting = setting[list(setting.keys())[0]]

    print("\n####### ####### ####### ####### #######")
    for item in setting:
        print(item, setting[item])
    print("####### ####### ####### ####### #######\n")
    # quit()

    # file_name = setting["file_name"]
    # use_intermediate = setting["use_intermediate"]
    # intermediate_recipe = setting["intermediate_recipe"]
    # N = setting["N"]

    particles = [i for i in setting["branches"].keys() if "DAUGHTER" in i]
    mother = "MOTHER"
    intermediates = [i for i in setting["branches"].keys() if "INTERMEDIATE" in i]
    intermediate_recipes = {}

    for idx, intermediate_name in enumerate(setting["intermediate_names"]):
        recipe = setting["intermediate_daughters_recursive"][intermediate_name]
        recipe_DAUGHTERS = []
        for item in recipe:
            for daughter_idx, code in enumerate(
                setting["particles"].values()
            ):  # for name, age in dictionary.iteritems():  (for Python 2.x)
                if code == item:
                    recipe_DAUGHTERS.append(f"DAUGHTER{daughter_idx + 1}")
        intermediate_recipes[f"INTERMEDIATE{idx + 1}"] = recipe_DAUGHTERS

    print("Opening...")
    file = uproot.open(f"{file_name}:DecayTree")
    branches = file.keys()
    branches = [a for a in branches if "_COV_" not in a]
    print("Cut _COV_ branches")
    if head != -1:
        events = file.arrays(branches, library="pd", entry_stop=head)
    else:
        events = file.arrays(branches, library="pd")

    print("Opened file as pd array")
    print(events.shape)

    print("Shuffling...")
    events = events.sample(frac=1)
    print("Shuffled")

    # print(events)
    for particle in particles:
        # Cut tuple to only keep muons, electrons, pions and kaons (for now)
        # print(np.unique(np.abs(events[f"{particle}_TRUEID"])))
        events = events[np.abs(events[f"{particle}_TRUEID"]).isin(pid_list)]

    # print(events)
    # quit()
    events = events.reset_index(drop=True)
    events = add_conditions(
        events, particles, mother, intermediates, intermediate_recipes
    )

    # cut_condition = "(MOTHER_TRUEID != 0) & (MOTHER_BKGCAT < 60)"
    # events = events.query(cut_condition)
    # print(events)

    tt.write_df_to_root(events, f"{file_name[:-5]}_more_vars.root")



import lhcb_rex.settings.globals as myGlobals
import torch
import json

from lhcb_rex.models.hetero_graph_based_GAN import (
    gGAN as hetero_gGAN,
)
from lhcb_rex.models.hetero_graph_based_diffusion_model import (
    diffusion_model,
)
from lhcb_rex.models.hetero_graph_based_GAN_smearing import (
    gGAN as hetero_gGAN_smearing,
)
from lhcb_rex.models.hetero_graph_based_diffusion_model_smearing import (
    diffusion_model_smearing,
)
from lhcb_rex.models.hetero_graph_based_VAE_smearing import (
    VAE_smearing,
)
from lhcb_rex.models.vanilla_GAN import (
    vanillaGAN,
)
from lhcb_rex.models.vanilla_diffusion import (
    vanillaDiffusion,
)
import importlib.resources

from pydantic import BaseModel, validator
from typing import Dict, Optional
import lhcb_rex.processing.transformers as tfs
import pickle
import time
from pathlib import Path
from lhcb_rex.tools.get_weights import HuggingFacePath


class HeteroModelHandler(BaseModel):
    branch_options: Dict = None  # Field(default_factory=dict)
    network_option: str
    load_model: Optional[str] = None
    override_network_params: Optional[str] = None
    gan: Optional[object] = None  # Placeholder for model instance
    Transformers: Optional[dict] = None  # Placeholder for model instance
    model_parameters: Optional[str] = "default_model_parameters.json"
    silent: bool = False
    diffusion_model: str = None

    @validator("load_model")
    def validate_load_model(cls, value):
        if value and not (value.endswith(".pickle") or value.endswith(".pkl")):
            raise ValueError("load_model must be a string ending in .pickle or .pkl")
        return value

    @validator("override_network_params")
    def validate_override_network_params(cls, value):
        if value and not value.endswith(".json"):
            raise ValueError("override_network_params must be a string ending in .json")
        return value

    def __init__(self, **data):
        super().__init__(**data)

        if self.branch_options is None:
            self.branch_options = {}
            if "reco_vertex" in self.network_option:
                self.branch_options["mother_targets"] = myGlobals.mother_targets
                self.branch_options["track_targets"] = myGlobals.track_targets
                self.branch_options["mother_conditions"] = myGlobals.mother_conditions
                self.branch_options["track_conditions"] = myGlobals.track_conditions
                self.branch_options["intermediate_targets"] = (
                    myGlobals.intermediate_targets
                )
                self.branch_options["intermediate_conditions"] = (
                    myGlobals.intermediate_conditions
                )
            elif "mom_smear" in self.network_option:
                self.branch_options["track_targets"] = myGlobals.smearing_track_targets
                self.branch_options["track_conditions"] = (
                    myGlobals.smearing_track_conditions
                )
                self.branch_options["edge_conditions"] = (
                    myGlobals.smearing_edge_conditions
                )
            elif "PID_trig" in self.network_option:
                self.branch_options["track_targets"] = myGlobals.PID_track_targets
                self.branch_options["track_conditions"] = myGlobals.PID_track_conditions
                self.branch_options["edge_conditions"] = myGlobals.PID_edge_conditions
            else:
                raise ValueError(
                    f"network_option={self.network_option} not implemented."
                )

        if self.model_parameters is not None:
            model_parameters = self.model_parameters
        with (
            importlib.resources.files("lhcb_rex")
            .joinpath(f"settings/{model_parameters}")
            .open("r") as file
        ):
            network_params = json.load(file)
        if self.override_network_params:
            network_params = json.load(self.override_network_params)

        for option in network_params:
            self.branch_options[option] = network_params[option]

        for option in [
            "mother_targets",
            "intermediate_targets",
            "track_targets",
            "mother_conditions",
            "intermediate_conditions",
            "track_conditions",
        ]:
            try:
                if isinstance(self.branch_options[option], tuple):
                    self.branch_options[option] = self.branch_options[option][0]
            except Exception:
                pass

        if self.network_option in ["mom_smear", "PID_trig"]:
            # add_physics_dims = self.network_option == "mom_smear"
            add_physics_dims = False
            self.gan = hetero_gGAN_smearing(
                track_targets_dims=len(self.branch_options["track_targets"]),
                track_conditions_dims=len(self.branch_options["track_conditions"]),
                hidden_channels=self.branch_options["hidden_channels"],
                track_latent_dims=self.branch_options["track_latent_dims"],
                edge_conditions_dims=len(self.branch_options["edge_conditions"]),
                silent=self.silent,
                # add_MBD=True,
                add_MBD=False,
                add_physics_dims=add_physics_dims,
            )

        elif self.network_option in [
            "mom_smear_diffusion",
            "mom_smear_flowmatching",
            "PID_trig_diffusion",
            "PID_trig_flowmatching",
        ]:
            self.gan = diffusion_model_smearing(
                track_targets_dims=len(self.branch_options["track_targets"]),
                track_conditions_dims=len(self.branch_options["track_conditions"]),
                hidden_channels=self.branch_options["hidden_channels"],
                track_latent_dims=len(self.branch_options["track_targets"]),
                edge_conditions_dims=len(self.branch_options["edge_conditions"]),
                steps=int(self.branch_options["steps"]),
                silent=self.silent,
            )
        elif self.network_option in [
            "reco_vertex_diffusion",
            "reco_vertex_flowmatching",
        ]:
            self.gan = diffusion_model(
                mother_targets_dims=len(self.branch_options["mother_targets"]),
                intermediate_targets_dims=int(
                    len(self.branch_options["intermediate_targets"])
                ),
                track_targets_dims=len(self.branch_options["track_targets"]),
                mother_conditions_dims=len(self.branch_options["mother_conditions"]),
                intermediate_conditions_dims=int(
                    len(self.branch_options["intermediate_conditions"])
                ),
                track_conditions_dims=len(self.branch_options["track_conditions"]),
                hidden_channels=self.branch_options["hidden_channels"],
                mother_latent_dims=len(self.branch_options["mother_targets"]),
                track_latent_dims=len(self.branch_options["track_targets"]),
                intermediate_latent_dims=int(
                    len(self.branch_options["intermediate_targets"])
                ),
                steps=int(self.branch_options["steps"]),
                silent=self.silent,
            )
        elif self.network_option in ["mom_smear_VAE"]:
            self.gan = VAE_smearing(
                track_targets_dims=len(self.branch_options["track_targets"]),
                track_conditions_dims=len(self.branch_options["track_conditions"]),
                hidden_channels=self.branch_options["hidden_channels"],
                track_latent_dims=len(self.branch_options["track_targets"]),
                edge_conditions_dims=len(self.branch_options["edge_conditions"]),
                steps=int(self.branch_options["steps"]),
                silent=self.silent,
            )

        elif self.network_option in ["reco_vertex_diffusion_eGAN"]:
            self.gan = diffusion_model(
                mother_targets_dims=len(self.branch_options["mother_targets"]),
                intermediate_targets_dims=int(
                    len(self.branch_options["intermediate_targets"])
                ),
                track_targets_dims=len(self.branch_options["track_targets"]),
                mother_conditions_dims=len(self.branch_options["mother_conditions"]),
                intermediate_conditions_dims=int(
                    len(self.branch_options["intermediate_conditions"])
                ),
                track_conditions_dims=len(self.branch_options["track_conditions"]),
                hidden_channels=self.branch_options["hidden_channels"],
                mother_latent_dims=len(self.branch_options["mother_targets"]),
                track_latent_dims=len(self.branch_options["track_targets"]),
                intermediate_latent_dims=int(
                    len(self.branch_options["intermediate_targets"])
                ),
                steps=int(self.branch_options["steps"]),
                silent=self.silent,
                eGAN=True,
            )
            try:
                if torch.cuda.is_available():
                    checkpoint = torch.load(self.diffusion_model)
                else:
                    checkpoint = torch.load(
                        self.diffusion_model, map_location=torch.device("cpu")
                    )
            except Exception:
                print(
                    f"WARNING: model ({self.diffusion_model}) not found, waiting 45 seconds and trying again (could be in process of being saved)."
                )
                time.sleep(45)
                checkpoint = torch.load(self.diffusion_model)
            try:
                self.gan.diffuser.load_state_dict(
                    checkpoint["diffuser_state_dict"]
                )  # Load only generator weights
            except Exception as e:  # legacy models
                print(f"{e}: probably legacy model")
                self.gan.diffuser.load_state_dict(
                    checkpoint["generator_state_dict"]
                )  # Load only generator weights
            self.gan.diffuser.eval()
            print(f"Loaded diffuser from {self.diffusion_model}...")

        else:
            self.gan = hetero_gGAN(
                mother_targets_dims=len(self.branch_options["mother_targets"]),
                intermediate_targets_dims=int(
                    len(self.branch_options["intermediate_targets"])
                ),
                track_targets_dims=len(self.branch_options["track_targets"]),
                mother_conditions_dims=len(self.branch_options["mother_conditions"]),
                intermediate_conditions_dims=int(
                    len(self.branch_options["intermediate_conditions"])
                ),
                track_conditions_dims=len(self.branch_options["track_conditions"]),
                hidden_channels=self.branch_options["hidden_channels"],
                mother_latent_dims=self.branch_options["mother_latent_dims"],
                track_latent_dims=self.branch_options["track_latent_dims"],
                intermediate_latent_dims=self.branch_options[
                    "intermediate_latent_dims"
                ],
                silent=self.silent,
            )

        self.gan.to(myGlobals.device)

        if self.load_model is not None:
            self.load(self.load_model)

        transformer_quantiles = pickle.load(
            importlib.resources.files("lhcb_rex")
            .joinpath("processing/transfomer_quantiles.pkl")
            .open("rb")
        )
        min_maxes = pickle.load(
            importlib.resources.files("lhcb_rex")
            .joinpath("processing/min_maxes.pkl")
            .open("rb")
        )

        self.Transformers = {}
        for i, (key, quantiles) in enumerate(transformer_quantiles.items()):
            transformer_i = tfs.UpdatedTransformer(min_maxes=min_maxes)
            transformer_i.fit(quantiles, key)
            self.Transformers[key] = transformer_i

    def save(self, save_loc):
        if (
            "diffusion" in self.network_option or "flowmatching" in self.network_option
        ) and "GAN" not in self.network_option:
            state_dict = {}
            try:
                state_dict["EMAdiffuser_state_dict"] = (
                    self.gan.EMA_diffuser.ema_model.state_dict()
                )
            except Exception as e:
                print(f"EMA not found: {e}")
                pass
            state_dict["diffuser_state_dict"] = (
                self.gan.diffuser.state_dict()
            )  # Save only generator weights
            try:
                state_dict["encoder_state_dict"] = self.gan.encoder.state_dict()
            except Exception as e:
                print(f"Encoder not found {e}")
                pass
            state_dict["branch_options"] = self.branch_options
        elif "VAE" in self.network_option:
            state_dict = {
                "encoder_state_dict": self.gan.encoder.state_dict(),  # Save only generator weights
                "decoder_state_dict": self.gan.decoder.state_dict(),  # Save only generator weights
                "branch_options": self.branch_options,
            }
        else:
            state_dict = {
                # "model_state_dict": self.gan.state_dict(),  # Model's parameters
                "generator_state_dict": self.gan.generator.state_dict(),  # Save only generator weights
                "branch_options": self.branch_options,
            }
        torch.save(state_dict, save_loc)
        print(f"Model saved to {save_loc}")

    def load(self, load_loc, silent=False, ema=False):

        if isinstance(load_loc, HuggingFacePath):
            if torch.cuda.is_available():
                checkpoint = torch.load(load_loc.path)
            else:
                checkpoint = torch.load(load_loc.path, map_location=torch.device("cpu"))
            self.gan.diffuser.load_state_dict(checkpoint["model_state_dict"])
            return True

        if not Path(load_loc).exists():
            print(f"WARNING: Unable to load weights from {load_loc}")
            return False

        try:
            if torch.cuda.is_available():
                checkpoint = torch.load(load_loc)
            else:
                checkpoint = torch.load(load_loc, map_location=torch.device("cpu"))
        except Exception:
            print(
                f"WARNING: model ({load_loc}) not found, waiting 45 seconds and trying again (could be in process of being saved)."
            )
            time.sleep(45)
            checkpoint = torch.load(load_loc)

        try:  # for legacy models
            self.gan.load_state_dict(checkpoint["model_state_dict"])
        except Exception:
            if (
                "diffusion" in self.network_option
                or "flowmatching" in self.network_option
            ) and "GAN" not in self.network_option:
                if ema:
                    self.gan.diffuser.load_state_dict(
                        checkpoint["EMAdiffuser_state_dict"]
                    )  # Load only generator weights
                else:
                    self.gan.diffuser.load_state_dict(
                        checkpoint["diffuser_state_dict"]
                    )  # Load only generator weights
                try:
                    self.gan.encoder.load_state_dict(checkpoint["encoder_state_dict"])
                except Exception as e:
                    print(f"Unable to load encoder {e}")
                    pass
            elif "VAE" in self.network_option:
                self.gan.encoder.load_state_dict(checkpoint["encoder_state_dict"])
                self.gan.decoder.load_state_dict(checkpoint["decoder_state_dict"])
            else:
                self.gan.generator.load_state_dict(
                    checkpoint["generator_state_dict"]
                )  # Load only generator weights
        self.branch_options = checkpoint["branch_options"]
        if not silent:
            print(f"Model loaded from {load_loc}")
        return True


class ModelHandler:
    def __init__(
        self,
        branch_options={},
        graphify=True,
        hidden_channels=[256, 512, 256],
        graph_latent_dims=2,
        node_latent_dims=4,
        latent_dims=2,
        extra_latent_dims=0,
        beta=80.0,  # 60.
        lr=0.0001,
        N_graph_layers=3,
        use_batch_norm=False,
        use_dropout=False,
        load_loc=None,
        load_network_params=None,
        use_a_GAN=False,
        hypergraphs=False,
        silent=False,
        network_option="",
    ):
        self.silent = silent
        self.branch_options = branch_options
        if self.branch_options == {}:
            self.branch_options["targets_graph"] = myGlobals.smearPV_targets
            self.branch_options["conditions_graph"] = myGlobals.smearPV_conditions

        self.options = {
            "graphify": graphify,
            "hidden_channels": hidden_channels,
            "graph_latent_dims": graph_latent_dims,
            "node_latent_dims": node_latent_dims,
            "latent_dims": latent_dims,
            "extra_latent_dims": extra_latent_dims,
            "beta": beta,  # 60.
            "lr": lr,
            "N_graph_layers": N_graph_layers,
            "use_batch_norm": use_batch_norm,
            "use_dropout": use_dropout,
        }

        self.hidden_channels = self.options["hidden_channels"]
        self.graph_latent_dims = self.options["graph_latent_dims"]
        self.node_latent_dims = self.options["node_latent_dims"]
        self.N_graph_layers = self.options["N_graph_layers"]
        self.use_batch_norm = self.options["use_batch_norm"]
        self.use_dropout = self.options["use_dropout"]
        self.latent_dims = self.options["latent_dims"]
        self.extra_latent_dims = self.options["extra_latent_dims"]
        self.beta = self.options["beta"]
        self.lr = self.options["lr"]
        self.graphify = self.options["graphify"]

        self.loss_fn = torch.nn.MSELoss()

        if load_loc is not None:
            self.load_branch_options(load_loc)

        if load_network_params is not None:
            with open(load_network_params) as json_file:
                network_params = json.load(json_file)
            for option in network_params:
                self.options[option] = network_params[option]

        # print(self.options)
        self.options["targets_graph"] = self.branch_options["targets_graph"]
        self.options["conditions_graph"] = self.branch_options["conditions_graph"]

        if "diffusion" in network_option:
            self.gan = vanillaDiffusion(
                targets_dims=len(self.options["targets_graph"]),
                conditions_dims=len(self.options["conditions_graph"]),
                hidden_channels=self.options["hidden_channels"],
                latent_dims=self.extra_latent_dims,
                silent=self.silent,
            )

        else:
            self.gan = vanillaGAN(
                targets_dims=len(self.options["targets_graph"]),
                conditions_dims=len(self.options["conditions_graph"]),
                hidden_channels=self.options["hidden_channels"],
                latent_dims=self.options["latent_dims"] + self.extra_latent_dims,
                silent=self.silent,
            )

        self.optimizer = torch.optim.Adam(self.gan.parameters(), lr=self.lr)
        # self.gan.to(myGlobals.device)

        self.beta = self.options["beta"]
        self.graphify = self.options["graphify"]

        if load_loc is not None:
            self.load(load_loc)

    def save(self, save_loc):
        state_dict = {
            "model_state_dict": self.gan.state_dict(),  # Model's parameters
            "branch_options": self.branch_options,
        }
        torch.save(state_dict, save_loc)
        print(f"Model saved to {save_loc}")

    def load(self, load_loc, silent=False):

        if isinstance(load_loc, HuggingFacePath):
            if torch.cuda.is_available():
                checkpoint = torch.load(load_loc.path)
            else:
                checkpoint = torch.load(load_loc.path, map_location=torch.device("cpu"))
            self.gan.load_state_dict(checkpoint["model_state_dict"])
            return True

        if not Path(load_loc).exists():
            print(f"WARNING: Unable to load weights from {load_loc}")
            return False

        try:
            if torch.cuda.is_available():
                checkpoint = torch.load(load_loc)
            else:
                checkpoint = torch.load(load_loc, map_location=torch.device("cpu"))
        except Exception:
            print(
                f"WARNING: model ({load_loc}) not found, waiting 45 seconds and trying again (could be in process of being saved)."
            )
            time.sleep(45)
            checkpoint = torch.load(load_loc)

        self.gan.load_state_dict(checkpoint["model_state_dict"])
        self.branch_options = checkpoint["branch_options"]
        if not silent:
            print(f"Model loaded from {load_loc}")

        return True



import pkg_resources

DATA_PATH = pkg_resources.resource_filename("lhcb_rex", "")

from .inference.runner import run_from_tuple
from .inference.runner import run

__all__ = [
    "run",
    "run_from_tuple",
]  # controls what is imported if someone were to from <module> import *



import uproot3
import numpy as np


def write_df_to_root(df, output_name):
    print(f"Writing {output_name}...")
    branch_dict = {}
    data_dict = {}
    dtypes = df.dtypes
    used_columns = []  # stop repeat columns, kpipi_correction was getting repeated
    for dtype, branch in enumerate(df.keys()):
        if branch not in used_columns:
            if dtypes[dtype] == "uint32":
                dtypes[dtype] = "int32"
            if dtypes[dtype] == "uint64":
                dtypes[dtype] = "int64"
            branch_dict[branch] = dtypes[dtype]
            # stop repeat columns, kpipi_correction was getting repeated
            if np.shape(df[branch].shape)[0] > 1:
                data_dict[branch] = df[branch].iloc[:, 0]
            else:
                data_dict[branch] = df[branch]
        used_columns.append(branch)
    with uproot3.recreate(output_name) as f:
        f["DecayTree"] = uproot3.newtree(branch_dict)
        f["DecayTree"].extend(data_dict)



import lhcb_rex.settings.globals as myGlobals

from torch_geometric.data import InMemoryDataset
import pickle
import os

from lhcb_rex.data.data_handler import (
    # DataHandler,
    # HyperDataHandler,
    HeteroDataHandler,
)
from lhcb_rex.data.data_handlerGENERALISED import (
    HeteroDataHandlerGENERALISED,
)
import os.path
import shutil
from pathlib import Path
import importlib.resources


class GENERALISED_HeteroGraphDataset(InMemoryDataset):
    path = ""
    splits = 1
    max_N_samples_per = -1

    def __init__(
        self,
        root,  # storage location
        config=None,
        force_reload=False,
        path="",  # path to root file
        mode="build_graphs",
        mother_targets=[],
        intermediate_targets=[],
        track_targets=[],
        mother_conditions=[],
        intermediate_conditions=[],
        track_conditions=[],
        edge_conditions=[],
        use_weights=True,
        splits=1,
        processID=-1,
        smearingnet=False,
        PIDnet=False,
        require_validation_variables=False,
    ):
        if config:
            with open(config, "rb") as handle:
                self.setting = pickle.load(handle)
                self.setting = self.setting[list(self.setting.keys())[0]]

            print("\n####### ####### ####### ####### #######")
            for item in self.setting:
                print(item, self.setting[item])
            print("####### ####### ####### ####### #######\n")
            # quit()

        if smearingnet or PIDnet:
            myGlobals.personalised_track_node_types = True

        GENERALISED_HeteroGraphDataset.path = path

        GENERALISED_HeteroGraphDataset.mother_targets = mother_targets
        GENERALISED_HeteroGraphDataset.intermediate_targets = intermediate_targets
        GENERALISED_HeteroGraphDataset.track_targets = track_targets
        GENERALISED_HeteroGraphDataset.mother_conditions = mother_conditions
        GENERALISED_HeteroGraphDataset.intermediate_conditions = intermediate_conditions
        GENERALISED_HeteroGraphDataset.track_conditions = track_conditions
        GENERALISED_HeteroGraphDataset.edge_conditions = edge_conditions

        GENERALISED_HeteroGraphDataset.splits = splits

        GENERALISED_HeteroGraphDataset.processID = processID

        GENERALISED_HeteroGraphDataset.root = root
        GENERALISED_HeteroGraphDataset.smearingnet = smearingnet
        GENERALISED_HeteroGraphDataset.PIDnet = PIDnet
        GENERALISED_HeteroGraphDataset.require_validation_variables = (
            require_validation_variables
        )

        self.use_weights = use_weights
        self.split_idx = 0
        self.splits = splits

        super().__init__(root, force_reload=force_reload)

        self.mode = mode

        print("self.processed_paths[2]:", self.processed_paths[2])
        filehandler = open(self.processed_paths[2], "rb")
        self.options = pickle.load(filehandler)

        if os.path.isfile(f"{self.processed_paths[0]}"):
            pass
        else:
            print(
                f"{self.processed_paths[0]} doesnt exist: fail, adding +1 to split_idx"
            )
            self.load_next()

        if mode == "train" or mode == "build_graphs":
            self.load(self.processed_paths[0])
        if mode == "test" or mode == "build_graphs":
            self.load(self.processed_paths[1])

        if (mode == "train" or mode == "build_graphs") and self.use_weights:
            filehandler = open(self.processed_paths[3], "rb")
            weights = pickle.load(filehandler)
            self.options["training_weights"] = weights

        try:
            if mode == "train" or mode == "build_graphs":
                filehandler = open(self.processed_paths[-2], "rb")
                fully_reco_bool = pickle.load(filehandler)
                self.options["fully_reco_bool"] = fully_reco_bool
            if mode == "test" or mode == "build_graphs":
                filehandler = open(self.processed_paths[-1], "rb")
                fully_reco_bool = pickle.load(filehandler)
                self.options["fully_reco_bool"] = fully_reco_bool
        except Exception:
            pass

        my_file = Path(
            self.processed_paths[2].split("options.pt")[0] + "transfomer_quantiles.pkl"
        )
        if not my_file.is_file():
            print(
                "copying transformers to: ",
                self.processed_paths[2].split("options.pt")[0]
                + "transfomer_quantiles.pkl",
            )
            transformers = importlib.resources.files("lhcb_rex").joinpath(
                "processing/transfomer_quantiles.pkl"
            )
            shutil.copyfile(
                transformers,
                self.processed_paths[2].split("options.pt")[0]
                + "transfomer_quantiles.pkl",
            )

        my_file = Path(self.processed_paths[2].split("options.pt")[0] + "min_maxes.pkl")
        if not my_file.is_file():
            print(
                "copying min_maxes to: ",
                self.processed_paths[2].split("options.pt")[0] + "min_maxes.pkl",
            )
            min_maxes = importlib.resources.files("lhcb_rex").joinpath(
                "processing/min_maxes.pkl"
            )
            shutil.copyfile(
                min_maxes,
                self.processed_paths[2].split("options.pt")[0] + "min_maxes.pkl",
            )

    def get_paths(self, dir_path):
        while 1 == 1:
            if self.split_idx >= self.splits - 1:
                self.split_idx = 0
            processed_paths_i = [
                (
                    file.replace(".pt", f"_{self.split_idx}.pt")
                    if file != "options.pt"
                    else file
                )
                for file in self.files
            ]
            if os.path.isfile(f"{dir_path}/{processed_paths_i[0]}"):
                break
            else:
                print(
                    f"{dir_path}/{processed_paths_i[0]} doesnt exist: fail, adding +1 to split_idx"
                )
                self.split_idx += 1

        print(processed_paths_i)
        return processed_paths_i

    def load_next(self):
        self.split_idx += 1
        if self.split_idx >= self.splits - 1:
            self.split_idx = 0
        print(f"Loading split {self.split_idx} of {self.splits}...")
        dir_path = os.path.dirname(self.processed_paths[0])

        processed_paths_i = self.get_paths(dir_path)

        if self.mode == "train" or self.mode == "build_graphs":
            self.load(f"{dir_path}/{processed_paths_i[0]}")
        if self.mode == "test" or self.mode == "build_graphs":
            self.load(f"{dir_path}/{processed_paths_i[1]}")

        if (self.mode == "train" or self.mode == "build_graphs") and self.use_weights:
            filehandler = open(f"{dir_path}/{processed_paths_i[3]}", "rb")
            weights = pickle.load(filehandler)
            self.options["training_weights"] = weights

        try:
            if self.mode == "train" or self.mode == "build_graphs":
                filehandler = open(f"{dir_path}/{processed_paths_i[-2]}", "rb")
                fully_reco_bool = pickle.load(filehandler)
                self.options["fully_reco_bool"] = fully_reco_bool

            if self.mode == "test" or self.mode == "build_graphs":
                filehandler = open(f"{dir_path}/{processed_paths_i[-1]}", "rb")
                fully_reco_bool = pickle.load(filehandler)
                self.options["fully_reco_bool"] = fully_reco_bool
        except Exception:
            pass

        # self.processed_paths = processed_paths_i

    @property
    def processed_file_names(self):
        self.files = [
            "data.pt",
            "data_val.pt",
            "options.pt",
        ]
        if self.use_weights:
            self.files.append("train_weights.pt")
        self.files.append("fully_reco_bool_train.pt")
        self.files.append("fully_reco_bool_test.pt")

        paths = [
            file.replace(".pt", "_0.pt") if file != "options.pt" else file
            for file in self.files
        ]
        return paths

    def process(self):
        data_handler = HeteroDataHandlerGENERALISED(
            config=self.setting,
            path=GENERALISED_HeteroGraphDataset.path,
            N=int(GENERALISED_HeteroGraphDataset.max_N_samples_per),
            mother_targets=GENERALISED_HeteroGraphDataset.mother_targets,
            intermediate_targets=GENERALISED_HeteroGraphDataset.intermediate_targets,
            track_targets=GENERALISED_HeteroGraphDataset.track_targets,
            mother_conditions=GENERALISED_HeteroGraphDataset.mother_conditions,
            intermediate_conditions=GENERALISED_HeteroGraphDataset.intermediate_conditions,
            track_conditions=GENERALISED_HeteroGraphDataset.track_conditions,
            edge_conditions=GENERALISED_HeteroGraphDataset.edge_conditions,
            testing_frac=0.1,
            use_weights=self.use_weights,
            splits=GENERALISED_HeteroGraphDataset.splits,
            processID=GENERALISED_HeteroGraphDataset.processID,
            smearingnet=GENERALISED_HeteroGraphDataset.smearingnet,
            PIDnet=GENERALISED_HeteroGraphDataset.PIDnet,
            require_validation_variables=GENERALISED_HeteroGraphDataset.require_validation_variables,
        )
        # if HeteroGraphDataset.processID == -1:
        # for split in range(HeteroGraphDataset.splits):

        split = GENERALISED_HeteroGraphDataset.processID

        print(f"split number (process): {split}")

        # data_handler.make_split(split)
        print(self.processed_paths[0].replace("_0.pt", f"_{split}.pt"))

        if self.use_weights:
            weights_i = data_handler.data_train_list_weights
            filehandler = open(
                self.processed_paths[3].replace("_0.pt", f"_{split}.pt"), "wb"
            )
            pickle.dump(weights_i, filehandler)

        self.save(
            data_handler.data_train,
            self.processed_paths[0].replace("_0.pt", f"_{split}.pt"),
        )
        self.save(
            data_handler.data_test,
            self.processed_paths[1].replace("_0.pt", f"_{split}.pt"),
        )

        fully_reco_bool_train = data_handler.data_train_list_full_reco
        fully_reco_bool_test = data_handler.data_test_list_full_reco
        filehandler = open(
            self.processed_paths[-2].replace("_0.pt", f"_{split}.pt"), "wb"
        )
        pickle.dump(fully_reco_bool_train, filehandler)
        filehandler = open(
            self.processed_paths[-1].replace("_0.pt", f"_{split}.pt"), "wb"
        )
        pickle.dump(fully_reco_bool_test, filehandler)

        options = {}
        options["path"] = GENERALISED_HeteroGraphDataset.path
        # options["particles_involved"] = GENERALISED_HeteroGraphDataset.particles_involved
        options["settings"] = self.setting

        options["mother_targets"] = (GENERALISED_HeteroGraphDataset.mother_targets,)
        options["intermediate_targets"] = (
            GENERALISED_HeteroGraphDataset.intermediate_targets,
        )
        options["track_targets"] = (GENERALISED_HeteroGraphDataset.track_targets,)
        options["mother_conditions"] = (
            GENERALISED_HeteroGraphDataset.mother_conditions,
        )
        options["intermediate_conditions"] = (
            GENERALISED_HeteroGraphDataset.intermediate_conditions,
        )
        options["track_conditions"] = (GENERALISED_HeteroGraphDataset.track_conditions,)
        options["edge_conditions"] = (GENERALISED_HeteroGraphDataset.edge_conditions,)

        filehandler = open(self.processed_paths[2], "wb")
        pickle.dump(options, filehandler)


class HeteroGraphDataset(InMemoryDataset):
    particles_involved = [1, 2]
    fully_reco = False
    max_N_samples_per = 1000
    path = ""
    splits = 1

    def __init__(
        self,
        root,  # storage location
        max_N_samples_per=-1,  # limit to number of graphs to create
        fully_reco=False,
        force_reload=False,
        path="",  # path to root file
        particles_involved=[1, 2],
        intermediates=[],
        mode="build_graphs",
        mother_targets=[],
        intermediate_targets=[],
        track_targets=[],
        mother_conditions=[],
        intermediate_conditions=[],
        track_conditions=[],
        edge_conditions=[],
        use_weights=True,
        splits=1,
        mother_N=3,
        intermediate_N=[],
        processID=-1,
        smearingnet=False,
        PIDnet=False,
        require_validation_variables=False,
    ):
        if smearingnet or PIDnet:
            myGlobals.personalised_track_node_types = True

        HeteroGraphDataset.particles_involved = particles_involved
        HeteroGraphDataset.intermediates = intermediates
        HeteroGraphDataset.N = len(particles_involved)
        HeteroGraphDataset.max_N_samples_per = max_N_samples_per
        HeteroGraphDataset.fully_reco = fully_reco
        HeteroGraphDataset.path = path

        HeteroGraphDataset.mother_targets = mother_targets
        HeteroGraphDataset.intermediate_targets = intermediate_targets
        HeteroGraphDataset.track_targets = track_targets
        HeteroGraphDataset.mother_conditions = mother_conditions
        HeteroGraphDataset.intermediate_conditions = intermediate_conditions
        HeteroGraphDataset.track_conditions = track_conditions
        HeteroGraphDataset.edge_conditions = edge_conditions

        HeteroGraphDataset.splits = splits
        HeteroGraphDataset.mother_N = mother_N
        HeteroGraphDataset.intermediate_N = intermediate_N

        HeteroGraphDataset.processID = processID

        HeteroGraphDataset.root = root
        HeteroGraphDataset.smearingnet = smearingnet
        HeteroGraphDataset.PIDnet = PIDnet
        HeteroGraphDataset.require_validation_variables = require_validation_variables

        self.use_weights = use_weights
        self.split_idx = 0
        self.splits = splits

        super().__init__(root, force_reload=force_reload)

        self.mode = mode

        filehandler = open(self.processed_paths[2], "rb")
        self.options = pickle.load(filehandler)

        if os.path.isfile(f"{self.processed_paths[0]}"):
            pass
        else:
            print(
                f"{self.processed_paths[0]} doesnt exist: fail, adding +1 to split_idx"
            )
            self.load_next()

        if mode == "train" or mode == "build_graphs":
            self.load(self.processed_paths[0])
        if mode == "test" or mode == "build_graphs":
            self.load(self.processed_paths[1])

        if (mode == "train" or mode == "build_graphs") and self.use_weights:
            filehandler = open(self.processed_paths[3], "rb")
            weights = pickle.load(filehandler)
            self.options["training_weights"] = weights

        try:
            if mode == "train" or mode == "build_graphs":
                filehandler = open(self.processed_paths[-2], "rb")
                fully_reco_bool = pickle.load(filehandler)
                self.options["fully_reco_bool"] = fully_reco_bool
            if mode == "test" or mode == "build_graphs":
                filehandler = open(self.processed_paths[-1], "rb")
                fully_reco_bool = pickle.load(filehandler)
                self.options["fully_reco_bool"] = fully_reco_bool
        except Exception:
            pass

    def get_paths(self, dir_path):
        while 1 == 1:
            if self.split_idx >= self.splits - 1:
                self.split_idx = 0
            processed_paths_i = [
                (
                    file.replace(".pt", f"_{self.split_idx}.pt")
                    if file != "options.pt"
                    else file
                )
                for file in self.files
            ]
            if os.path.isfile(f"{dir_path}/{processed_paths_i[0]}"):
                break
            else:
                print(
                    f"{dir_path}/{processed_paths_i[0]} doesnt exist: fail, adding +1 to split_idx"
                )
                self.split_idx += 1

        print(processed_paths_i)
        return processed_paths_i

    def load_next(self):
        self.split_idx += 1
        if self.split_idx >= self.splits - 1:
            self.split_idx = 0
        print(f"Loading split {self.split_idx} of {self.splits}...")
        dir_path = os.path.dirname(self.processed_paths[0])

        processed_paths_i = self.get_paths(dir_path)

        if self.mode == "train" or self.mode == "build_graphs":
            self.load(f"{dir_path}/{processed_paths_i[0]}")
        if self.mode == "test" or self.mode == "build_graphs":
            self.load(f"{dir_path}/{processed_paths_i[1]}")

        if self.mode == "train" or self.mode == "build_graphs":
            filehandler = open(f"{dir_path}/{processed_paths_i[3]}", "rb")
            weights = pickle.load(filehandler)
            self.options["training_weights"] = weights

        try:
            if self.mode == "train" or self.mode == "build_graphs":
                filehandler = open(f"{dir_path}/{processed_paths_i[-2]}", "rb")
                fully_reco_bool = pickle.load(filehandler)
                self.options["fully_reco_bool"] = fully_reco_bool

            if self.mode == "test" or self.mode == "build_graphs":
                filehandler = open(f"{dir_path}/{processed_paths_i[-1]}", "rb")
                fully_reco_bool = pickle.load(filehandler)
                self.options["fully_reco_bool"] = fully_reco_bool
        except Exception:
            pass

    @property
    def processed_file_names(self):
        self.files = [
            "data.pt",
            "data_val.pt",
            "options.pt",
            # "train_weights.pt",
            "fully_reco_bool_train.pt",
            "fully_reco_bool_test.pt",
        ]
        if self.use_weights:
            self.files.append("train_weights.pt")

        # if HeteroGraphDataset.splits == 1:
        #     return self.files
        # else:
        # return ["data_0.pt", "data_val_0.pt", "options.pt", "train_weights_0.pt"]

        paths = [
            file.replace(".pt", "_0.pt") if file != "options.pt" else file
            for file in self.files
        ]
        # path_there = []
        # for path in paths:
        #     print(f'path: {path} ',os.path.isfile(f'{HeteroGraphDataset.root}/processed/{path}'))
        #     path_there.append(os.path.isfile(f'{HeteroGraphDataset.root}/processed/{path}'))
        # if False in path_there:
        #     raise ValueError("Missing file")

        return paths

    def process(self):
        data_handler = HeteroDataHandler(
            particles_involved=HeteroGraphDataset.particles_involved,
            intermediates=HeteroGraphDataset.intermediates,
            path=HeteroGraphDataset.path,
            N=int(HeteroGraphDataset.max_N_samples_per),
            mother_targets=HeteroGraphDataset.mother_targets,
            intermediate_targets=HeteroGraphDataset.intermediate_targets,
            track_targets=HeteroGraphDataset.track_targets,
            mother_conditions=HeteroGraphDataset.mother_conditions,
            intermediate_conditions=HeteroGraphDataset.intermediate_conditions,
            track_conditions=HeteroGraphDataset.track_conditions,
            edge_conditions=HeteroGraphDataset.edge_conditions,
            graphify=True,
            testing_frac=0.1,
            # cut=cut,
            use_weights=self.use_weights,
            splits=HeteroGraphDataset.splits,
            mother_N=HeteroGraphDataset.mother_N,
            intermediate_N=HeteroGraphDataset.intermediate_N,
            processID=HeteroGraphDataset.processID,
            smearingnet=HeteroGraphDataset.smearingnet,
            PIDnet=HeteroGraphDataset.PIDnet,
            require_validation_variables=HeteroGraphDataset.require_validation_variables,
        )

        # if HeteroGraphDataset.processID == -1:
        # for split in range(HeteroGraphDataset.splits):

        split = HeteroGraphDataset.processID

        print(f"split number (process): {split}")

        # data_handler.make_split(split)
        print(self.processed_paths[0].replace("_0.pt", f"_{split}.pt"))

        if self.use_weights:
            weights_i = data_handler.data_train_list_weights
            filehandler = open(
                self.processed_paths[3].replace("_0.pt", f"_{split}.pt"), "wb"
            )
            pickle.dump(weights_i, filehandler)

        self.save(
            data_handler.data_train,
            self.processed_paths[0].replace("_0.pt", f"_{split}.pt"),
        )
        self.save(
            data_handler.data_test,
            self.processed_paths[1].replace("_0.pt", f"_{split}.pt"),
        )

        fully_reco_bool_train = data_handler.data_train_list_full_reco
        fully_reco_bool_test = data_handler.data_test_list_full_reco
        filehandler = open(
            self.processed_paths[-2].replace("_0.pt", f"_{split}.pt"), "wb"
        )
        pickle.dump(fully_reco_bool_train, filehandler)
        filehandler = open(
            self.processed_paths[-1].replace("_0.pt", f"_{split}.pt"), "wb"
        )
        pickle.dump(fully_reco_bool_test, filehandler)

        options = {}
        options["path"] = HeteroGraphDataset.path
        options["particles_involved"] = HeteroGraphDataset.particles_involved
        options["intermediates"] = HeteroGraphDataset.intermediates

        options["mother_targets"] = (HeteroGraphDataset.mother_targets,)
        options["intermediate_targets"] = (HeteroGraphDataset.intermediate_targets,)
        options["track_targets"] = (HeteroGraphDataset.track_targets,)
        options["mother_conditions"] = (HeteroGraphDataset.mother_conditions,)
        options["intermediate_conditions"] = (
            HeteroGraphDataset.intermediate_conditions,
        )
        options["track_conditions"] = (HeteroGraphDataset.track_conditions,)
        options["edge_conditions"] = (HeteroGraphDataset.edge_conditions,)

        filehandler = open(self.processed_paths[2], "wb")
        pickle.dump(options, filehandler)


# class GraphDataset(InMemoryDataset):

#     particles_involved = [1, 2]
#     fully_reco = False
#     max_N_samples_per = 1000
#     path = ""

#     def __init__(
#         self,
#         root,  # storage location
#         max_N_samples_per=-1,  # limit to number of graphs to create
#         fully_reco=False,
#         force_reload=False,
#         path="",  # path to root file
#         particles_involved=[1, 2],
#         mode="build_graphs",
#         targets_graph=[],
#         targets_node=[],
#         conditions_graph=[],
#         conditions_node=[],
#         conditions_edge=[],
#         use_weights=True,
#     ):

#         GraphDataset.particles_involved = particles_involved
#         GraphDataset.N = len(particles_involved)
#         GraphDataset.max_N_samples_per = max_N_samples_per
#         GraphDataset.fully_reco = fully_reco
#         GraphDataset.path = path
#         GraphDataset.targets_graph = targets_graph
#         GraphDataset.targets_node = targets_node
#         GraphDataset.conditions_graph = conditions_graph
#         GraphDataset.conditions_node = conditions_node
#         GraphDataset.conditions_edge = conditions_edge

#         self.use_weights = use_weights

#         super().__init__(root, force_reload=force_reload)

#         if mode == "train" or mode == "build_graphs":
#             self.load(self.processed_paths[0])
#         if mode == "test" or mode == "build_graphs":
#             self.load(self.processed_paths[1])

#         filehandler = open(self.processed_paths[2], "rb")
#         self.options = pickle.load(filehandler)

#         # # Shorten
#         # N_nodes = int(np.shape(self.data.targets)[0]/np.shape(self.data.N_daughters)[0])
#         # if N_nodes == 2: N_edges = 2
#         # if N_nodes == 3: N_edges = 6
#         # if N_nodes == 4:  N_edges = 12
#         # N_samples = np.shape(self.data.N_daughters)[0]
#         # if max_N_samples_per != -1:
#         #     if N_samples < max_N_samples_per:
#         #         pass
#         #     else:

#         #         self.data.edge_index = self.data.edge_index[:, :N_edges * max_N_samples_per]
#         #         self.data.targets = self.data.targets[:N_nodes * max_N_samples_per, :]
#         #         self.data.conditions = self.data.conditions[:N_nodes * max_N_samples_per, :]
#         #         self.data.node_targets = self.data.node_targets[:N_nodes * max_N_samples_per, :]
#         #         self.data.graph_targets = self.data.graph_targets[:max_N_samples_per, :]
#         #         self.data.N_daughters = self.data.N_daughters[:max_N_samples_per]
#         #         self.data.particles_involved = self.data.particles_involved[:max_N_samples_per]

#         #         # for a test:
#         #         # self.data.node_targets[:,1:] = torch.randn(self.data.node_targets[:, 1:].shape)/4.

#         #         print(f"Loaded only the first {max_N_samples_per} of {N_samples} graphs")

#         #         # Truncate slices
#         #         for key, slice_tensor in self.slices.items():
#         #             if key == 'edge_index':
#         #                 self.slices[key] = slice_tensor[:max_N_samples_per + 1]  # Include boundary
#         #             elif key in ['targets', 'conditions', 'node_targets']:
#         #                 self.slices[key] = slice_tensor[:N_nodes * max_N_samples_per + 1]
#         #             elif key in ['graph_targets', 'N_daughters', 'particles_involved']:
#         #                 self.slices[key] = slice_tensor[:max_N_samples_per + 1]

#     @property
#     def processed_file_names(self):
#         return ["data.pt", "data_val.pt", "options.pt"]

#     def process(self):

#         data_list = []
#         data_list_val = []

#         if GraphDataset.fully_reco:
#             cut = "fully_reco>0.5"
#         else:
#             cut = "fully_reco<0.5"

#         data_handler = DataHandler(
#             particles_involved=GraphDataset.particles_involved,
#             path=GraphDataset.path,
#             N=int(GraphDataset.max_N_samples_per),
#             targets_graph=GraphDataset.targets_graph,
#             targets_node=GraphDataset.targets_node,
#             conditions_graph=GraphDataset.conditions_graph,
#             conditions_node=GraphDataset.conditions_node,
#             conditions_edge=GraphDataset.conditions_edge,
#             graphify=True,
#             testing_frac=0.1,
#             cut=cut,
#             use_weights=self.use_weights,
#         )
#         data_list_i = data_handler.data_train
#         data_list += data_list_i

#         if self.use_weights:
#             weights = data_handler.data_train_list_weights

#         data_list_val_i = data_handler.data_test
#         data_list_val += data_list_val_i

#         self.save(data_list, self.processed_paths[0])
#         self.save(data_list_val, self.processed_paths[1])

#         options = {}
#         options["path"] = GraphDataset.path
#         options["particles_involved"] = GraphDataset.particles_involved
#         options["targets_graph"] = GraphDataset.targets_graph
#         options["targets_node"] = GraphDataset.targets_node
#         options["conditions_graph"] = GraphDataset.conditions_graph
#         options["conditions_node"] = GraphDataset.conditions_node
#         options["conditions_edge"] = GraphDataset.conditions_edge
#         options["training_weights"] = weights

#         filehandler = open(self.processed_paths[2], "wb")
#         pickle.dump(options, filehandler)



import lhcb_rex.settings.globals as myGlobals
import pandas as pd
import numpy as np
import lhcb_rex.processing.transformers as tfs
import pickle
import re
import uproot
import importlib.resources


class NoneError(Exception):
    pass


class Dataset:
    def __init__(self, particles_involved, filename, transformers=None):
        self.Transformers = transformers

        transformer_quantiles = pickle.load(
            importlib.resources.files("lhcb_rex")
            .joinpath("processing/transfomer_quantiles.pkl")
            .open("rb")
        )
        min_maxes = pickle.load(
            importlib.resources.files("lhcb_rex")
            .joinpath("processing/min_maxes.pkl")
            .open("rb")
        )

        self.Transformers = {}
        for i, (key, quantiles) in enumerate(transformer_quantiles.items()):
            transformer_i = tfs.UpdatedTransformer(min_maxes=min_maxes)
            transformer_i.fit(quantiles, key)
            self.Transformers[key] = transformer_i

        self.all_data = {"processed": None, "physical": None}

        self.filename = filename

        self.mother_particle = "MOTHER"
        self.daughter_particles = []
        for idx in particles_involved:
            self.daughter_particles.append(f"DAUGHTER{idx}")

    def pre_process(self, physical_data):
        df = {}

        if self.Transformers is None:
            self.Transformers = {}

        for column in list(physical_data.keys()):
            if column[:4] == "edge":
                transformer_column = re.sub(
                    r"DAUGHTER\d+_DAUGHTER\d+", "DAUGHTERN_DAUGHTERN", column
                )
            else:
                transformer_column = re.sub(r"DAUGHTER\d+", "DAUGHTER1", column)
                # transformer_column = re.sub(r"INTERMEDIATE\d+", "INTERMEDIATE", transformer_column)
                transformer_column = re.sub(
                    r"INTERMEDIATE\d+", "MOTHER", transformer_column
                )
                transformer_column = re.sub(
                    r"delta_\d+_P", "delta_0_P", transformer_column
                )
                transformer_column = re.sub(
                    r"delta_\d+_PT", "delta_0_PT", transformer_column
                )

            # print(column, transformer_column)
            if (  # personalised transformers
                "residualfrac" in column
                or "_Preco_overP" in column
                or "delta_PX" in column
                or "delta_PY" in column
                or "delta_PZ" in column
                or re.search("_PID.", column)
                or re.search("_ProbNN.", column)
                or "TRACK_CHI2NDOF" in column
                or "TRACK_GhostProb" in column
            ) and myGlobals.personalised_track_node_types:
                print(f"using personalised_track_node_types {column}")
                physical_data_i = np.asarray(physical_data[column]).copy()
                pid_column = np.abs(
                    np.asarray(physical_data[f"{column[:9]}_TRUEID"]).copy()
                )

                try:
                    out = np.zeros_like(physical_data_i)
                    for particle_type in [11, 13, 211, 321, 2212]:
                        where = np.where(pid_column == particle_type)
                        out[where] = self.Transformers[
                            f"{transformer_column}_{particle_type}"
                        ].process(physical_data_i[where])
                    df[column] = out

                except Exception:
                    # print(f"FAIL, {column}")
                    pass
            else:
                print(f"not using personalised_track_node_types {column}")
                physical_data_i = np.asarray(physical_data[column]).copy()

                if (
                    column == "file"
                    or column == "pass_stripping"
                    or column == "training_weight"
                    or column == "weight"
                    or column == "in_training"
                    or "N_daughters" in column
                    or column == "fully_reco"
                    or column in myGlobals.validation_variables
                ):
                    df[column] = physical_data_i
                if "DAUGHTER" in column and "_TRUEID" in column:
                    df[column] = self.Transformers[transformer_column].process(
                        physical_data_i
                    )
                else:
                    try:
                        df[column] = self.Transformers[transformer_column].process(
                            physical_data_i
                        )
                    except Exception:
                        # print(f"FAIL, {column}")
                        pass

        return pd.DataFrame.from_dict(df)

    def produce_physics_variables(self):
        for particle_i in self.daughter_particles:
            self.all_data["physical"][f"{particle_i}_P"] = np.sqrt(
                self.all_data["physical"][f"{particle_i}_PX"] ** 2
                + self.all_data["physical"][f"{particle_i}_PY"] ** 2
                + self.all_data["physical"][f"{particle_i}_PZ"] ** 2
            )

            self.all_data["physical"][f"{particle_i}_PT"] = np.sqrt(
                self.all_data["physical"][f"{particle_i}_PX"] ** 2
                + self.all_data["physical"][f"{particle_i}_PY"] ** 2
            )

            self.all_data["physical"][f"{particle_i}_TRUE_P"] = np.sqrt(
                self.all_data["physical"][f"{particle_i}_TRUEP_X"] ** 2
                + self.all_data["physical"][f"{particle_i}_TRUEP_Y"] ** 2
                + self.all_data["physical"][f"{particle_i}_TRUEP_Z"] ** 2
            )

            self.all_data["physical"][f"{particle_i}_TRUEP_T"] = np.sqrt(
                self.all_data["physical"][f"{particle_i}_TRUEP_X"] ** 2
                + self.all_data["physical"][f"{particle_i}_TRUEP_Y"] ** 2
            )

            self.all_data["physical"][f"{particle_i}_eta"] = -np.log(
                np.tan(
                    np.arcsin(
                        self.all_data["physical"][f"{particle_i}_PT"]
                        / self.all_data["physical"][f"{particle_i}_P"]
                    )
                    / 2.0
                )
            )

            self.all_data["physical"][f"{particle_i}_eta_TRUE"] = -np.log(
                np.tan(
                    np.arcsin(
                        self.all_data["physical"][f"{particle_i}_TRUEP_T"]
                        / self.all_data["physical"][f"{particle_i}_TRUE_P"]
                    )
                    / 2.0
                )
            )

        self.all_data["physical"]["kFold"] = np.random.randint(
            low=0,
            high=9,
            size=np.shape(
                self.all_data["physical"][f"{self.daughter_particles[0]}_PX"]
            )[0],
        )

    def check_TRUE_branches(self):
        for particle in self.daughter_particles + [self.mother_particle]:
            for dim in ["X", "Y", "Z"]:
                if f"{particle}_TRUEP_{dim}" not in list(
                    self.all_data["physical"].keys()
                ):
                    self.all_data["physical"][f"{particle}_TRUEP_{dim}"] = (
                        1000.0 * self.all_data["physical"][f"{particle}_P{dim}_TRUE"]
                    )
                    self.all_data["physical"][f"{particle}_P{dim}"] = (
                        1000.0 * self.all_data["physical"][f"{particle}_P{dim}"]
                    )
            if f"{particle}_TRUEID" not in list(self.all_data["physical"].keys()):
                self.all_data["physical"][f"{particle}_TRUEID"] = self.all_data[
                    "physical"
                ][f"{particle}_ID_TRUE"]
            if f"{particle}_mass" not in list(self.all_data["physical"].keys()):
                self.all_data["physical"][f"{particle}_mass"] = self.all_data[
                    "physical"
                ][f"{particle}_M"]

    def fill(self, data, testing_frac=0.1):
        if not isinstance(data, pd.DataFrame):
            raise NoneError("Dataset must be a pd.dataframe.")

        in_training = np.ones(data.shape[0])
        in_training[-int(testing_frac * data.shape[0]) :] = 0
        data["in_training"] = in_training

        self.all_data["physical"] = data

        ####
        # self.check_TRUE_branches()
        # self.produce_physics_variables()
        ####

        self.all_data["physical"] = self.all_data["physical"].loc[
            :, ~self.all_data["physical"].columns.str.contains("^Unnamed")
        ]

        self.all_data["processed"] = self.pre_process(self.all_data["physical"])

    def overwrite_branch_processed(self, data, branch):
        transformer_column = re.sub(r"DAUGHTER\d+", "DAUGHTER1", branch)
        # transformer_column = re.sub(r"INTERMEDIATE\d+", "INTERMEDIATE", transformer_column)
        transformer_column = re.sub(r"INTERMEDIATE\d+", "MOTHER", transformer_column)
        transformer_column = re.sub(r"delta_\d+_P", "delta_0_P", transformer_column)
        transformer_column = re.sub(r"delta_\d+_PT", "delta_0_PT", transformer_column)

        self.all_data["processed"][branch] = data
        self.all_data["physical"][branch] = self.Transformers[
            transformer_column
        ].unprocess(np.asarray(self.all_data["processed"][branch]).copy())

    def print_branches(self):
        for key in list(self.all_data["physical"].keys()):
            print(key)

    def get_branches(self, branches, processed=True, option=""):
        if not isinstance(branches, list):
            branches = [branches]

        if processed:
            missing = list(
                set(branches).difference(set(list(self.all_data["processed"].keys())))
            )
            branches = list(
                set(branches).intersection(set(list(self.all_data["processed"].keys())))
            )

            if len(missing) > 0:
                print(f"missing branches: {missing}\n {self.filename} \n")

            if option != "":
                output = self.all_data["processed"][branches + ["in_training"]]
            else:
                output = self.all_data["processed"][branches]

        else:
            missing = list(
                set(branches).difference(set(list(self.all_data["physical"].keys())))
            )
            branches = list(
                set(branches).intersection(set(list(self.all_data["physical"].keys())))
            )

            if len(missing) > 0:
                print(f"missing branches: {missing}\n {self.filename} \n")

            if option != "":
                output = self.all_data["physical"][branches + ["in_training"]]
            else:
                output = self.all_data["physical"][branches]

        if option == "training":
            output = output.query("in_training==1.0")
            output = output[branches]
        elif option == "testing":
            output = output.query("in_training==0.0")
            output = output[branches]

        return output

    def write_to_file(self, filename):
        print(f"Writting {filename}...")
        with uproot.recreate(filename) as new_file:
            new_file["DecayTree"] = self.all_data["physical"]
        print("Done.")



import lhcb_rex.settings.globals as myGlobals
from pydantic import BaseModel, validator
import lhcb_rex.models.model_handler as model_handler
from typing import Any
import lhcb_rex.processing.transformers as tfs
import pickle
import importlib.resources
import re
import torch
import pandas as pd
from rich import print
import numpy as np
import itertools
import lhcb_rex.tools.display as display
from diffusers import DDPMScheduler
from typing import Union
from lhcb_rex.tools.get_weights import HuggingFacePath

class DataContainer:
    def __init__(self):
        self.x = None
        self.data = {}

    def __setattr__(self, key, value):
        if key in ["x", "data"]:
            super().__setattr__(key, value)
        else:
            self.data[key] = value

    def __getattr__(self, key):
        if key in self.data:
            return self.data[key]
        raise AttributeError(f"'DataContainer' object has no attribute '{key}'")

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, key):
        return key in self.data


def batchify(edge, single_graph, batch_size, N_nodes=None):
    origin_N = N_nodes[edge[0]]
    destin_N = N_nodes[edge[-1]]

    batched_edges = []

    for i in range(batch_size):
        offset_origin = i * origin_N  # Shift nodes for each graph
        offset_destin = i * destin_N  # Shift nodes for each graph

        single_graph_offset = single_graph.clone()
        single_graph_offset[0] += offset_origin
        single_graph_offset[1] += offset_destin

        batched_edges.append(single_graph_offset)

    return torch.cat(batched_edges, dim=1)  # Concatenate along columns


class Network(BaseModel):
    network: str
    model_parameters: str = "default_model_parameters.json"
    pkl: Union[str, HuggingFacePath]
    Nparticles: int
    physical_units: str = "GeV"
    daughter_particle_names: list
    transformer_file: str = "transfomer_quantiles.pkl"
    EMA: bool = False
    N_diffusion_steps: int = 25

    Transformers: Any = None
    Model: Any = None
    conditions: Any = None
    edge_conditions: Any = None
    transformers: Any = None  # personalised
    edge_transformers: Any = None  # personalised
    query_network: Any = None
    targets: Any = None
    targets_full: Any = None

    mother_targets: Any = None
    intermediate_targets: Any = None
    track_targets: Any = None
    mother_conditions: Any = None
    intermediate_conditions: Any = None
    track_conditions: Any = None

    internal_daughter_particle_names: Any = None
    intermediate_particle_name: Any = None

    class Config:
        arbitrary_types_allowed = True

    @validator("model_parameters")
    def validate_model_parameters(cls, value):
        if value and not value.endswith(".json"):
            raise ValueError("model_parameters must be a string ending in .json")
        return value

    @validator("network")
    def validate_network(cls, value):
        network_options = [
            "PV_smear",
            "PV_smear_diffusion",
            "mom_smear",
            "PID_trig",
            "reco_vertex",
            "reco_vertex_diffusion",
            "mom_smear_diffusion",
            "PID_trig_diffusion",
            "mom_smear_flowmatching",
            "PID_trig_flowmatching",
            "reco_vertex_flowmatching",
        ]
        if value not in network_options:
            raise ValueError(f"network option must be one of: {network_options}")
        return value

    @validator("pkl")
    def validate_pkl(cls, value):
        if isinstance(value, HuggingFacePath): return value
        if not value.endswith(".pkl") and not value.endswith(".pickle"):
            raise ValueError("pkl should point to a pickle of network state")
        return value

    def rapidsimify_branches(self, branches):
        branches_prime = []
        for branch in branches:
            if match := re.search(r"_(TRUEP)(_?[A-Z]?)$", branch):
                replacement = f"_P{match.group(2).replace('_', '')}_TRUE"
                branches_prime.append(re.sub(r"_TRUEP(_?[A-Z]?)$", replacement, branch))
            elif match := re.search(r"_(TRUEE)(_?[A-Z]?)$", branch):
                replacement = f"_E{match.group(2).replace('_', '')}_TRUE"
                branches_prime.append(re.sub(r"_TRUEE(_?[A-Z]?)$", replacement, branch))
            elif match := re.search(r"_(TRUEENDVERTEX)(_?[A-Z]?)$", branch):
                replacement = f"_vtx{match.group(2).replace('_', '')}_TRUE"
                branches_prime.append(
                    re.sub(r"_(TRUEENDVERTEX)(_?[A-Z]?)$", replacement, branch)
                )
            elif match := re.search(r"_(TRUEORIGINVERTEX)(_?[A-Z]?)$", branch):
                replacement = f"_orig{match.group(2).replace('_', '')}_TRUE"
                branches_prime.append(
                    re.sub(r"_(TRUEORIGINVERTEX)(_?[A-Z]?)$", replacement, branch)
                )
            else:
                branches_prime.append(branch)
        return branches_prime

    def mom_branch_filter(self, branch):
        if re.search(r"delta_P[A-Z]", branch):
            return True
        if branch.endswith("TRUEP"):
            return True
        if branch.endswith("_P"):
            return True
        if branch.endswith("_PT"):
            return True
        if branch.endswith("_TRUEE"):
            return True
        if branch.endswith("_missing_P"):
            return True
        if branch.endswith("_missing_PT"):
            return True
        if re.search(r"TRUEP_[A-Z]", branch):
            return True
        if (
            re.search(r"_P[A-Z]", branch)
            and "PID" not in branch
            and "frac" not in branch
        ):
            return True
        if branch.endswith("_M"):
            return True
        if branch.endswith("_M_reco"):
            return True
        if branch.endswith("_mass"):
            return True
        if re.search(r"deltaEV_[A-Z]", branch):
            return True
        else:
            return False

    def load_transformers(self):
        transformer_quantiles = pickle.load(
            importlib.resources.files("lhcb_rex")
            .joinpath(f"processing/{self.transformer_file}")
            .open("rb")
        )
        min_maxes = pickle.load(
            importlib.resources.files("lhcb_rex")
            .joinpath("processing/min_maxes.pkl")
            .open("rb")
        )
        self.Transformers = {}
        for i, (key, quantiles) in enumerate(transformer_quantiles.items()):
            if "reco_vertex" in self.network or "PV_smear" in self.network:
                transformer_i = tfs.UpdatedTransformer(
                    min_maxes=min_maxes, use_min_max_for_mom_deltas=False
                )
            else:
                transformer_i = tfs.UpdatedTransformer(
                    min_maxes=min_maxes, use_min_max_for_mom_deltas=True
                )

            unit_converter = 1.0
            if self.physical_units == "GeV":
                if self.mom_branch_filter(key):
                    unit_converter = 1000.0  # 1000. means that this transformer (that was fit to MeV MC) will transform GeV branches - transformer(data*unit_converter)

            transformer_i.fit(quantiles, key, unit_converter=unit_converter)
            self.Transformers[key] = transformer_i

        return self.Transformers

    def initialise_PV_smearer(self):
        if "diffusion" in self.network:
            self.Model = model_handler.ModelHandler(
                graphify=False,
                hidden_channels=[256, 512, 256],
                latent_dims=len(myGlobals.smearPV_targets),
                extra_latent_dims=3,
                beta=80.0,
                lr=0.0001,
                silent=True,
                network_option=self.network,
            )
            success = self.Model.load(self.pkl, silent=True)
            if success:
                if isinstance(self.pkl, HuggingFacePath):
                    display.info_print(f"Model loaded from {self.pkl.path}")
                else:
                    display.info_print(f"Model loaded from {self.pkl}")
            self.Model.gan.diffuser.eval()
        else:
            self.Model = model_handler.ModelHandler(
                graphify=False,
                hidden_channels=[256, 512, 256],
                latent_dims=25,
                beta=80.0,
                lr=0.0001,
                silent=True,
                network_option=self.network,
            )
            success = self.Model.load(self.pkl, silent=True)
            if success:
                if isinstance(self.pkl, HuggingFacePath):
                    display.info_print(f"Model loaded from {self.pkl.path}")
                else:
                    display.info_print(f"Model loaded from {self.pkl}")
            self.Model.gan.generator.eval()
            self.Model.gan.discriminator.eval()

        self.transformers = self.Transformers

        targets_raw = self.Model.branch_options["targets_graph"]
        self.targets = self.rapidsimify_branches(targets_raw)
        for idx, target in enumerate(self.targets):
            self.transformers[target] = self.transformers[targets_raw[idx]]

        conditions_raw = self.Model.branch_options["conditions_graph"]
        self.conditions = self.rapidsimify_branches(conditions_raw)

        for idx, cond in enumerate(self.conditions):
            self.transformers[cond] = self.transformers[conditions_raw[idx]]
        self.query_network = self.query_network_vanilla

    def initialise_hetero_model(self):
        Model = model_handler.HeteroModelHandler(
            network_option=self.network,
            model_parameters=self.model_parameters,
            silent=True,
        )
        success = Model.load(self.pkl, silent=True, ema=self.EMA)
        if success:
            if isinstance(self.pkl, HuggingFacePath):
                display.info_print(f"Model loaded from {self.pkl.path}")
            else:
                display.info_print(f"Model loaded from {self.pkl}, ema:{self.EMA}")


        if "diffusion" not in self.network and "flowmatching" not in self.network:
            Model.gan.generator.eval()
            Model.gan.discriminator.eval()
        else:
            Model.gan.diffuser.eval()

        return Model

    def initialise_unique_node_hetero_model(self):
        self.targets = self.Model.branch_options["track_targets"]
        conditions_raw = self.Model.branch_options["track_conditions"]
        conditions = self.rapidsimify_branches(conditions_raw)
        self.Model.branch_options["edge_conditions"] = self.Model.branch_options[
            "edge_conditions"
        ][0]
        edge_conditions = self.Model.branch_options["edge_conditions"]
        if not isinstance(edge_conditions, list): 
            edge_conditions = [edge_conditions]

        self.transformers = self.Transformers

        self.conditions = []
        for particle in self.internal_daughter_particle_names:
            for cond in conditions:
                self.conditions.append(cond.replace("DAUGHTERN", particle))

        self.targets_full = []
        for particle in self.internal_daughter_particle_names:
            for targ in self.targets:
                self.targets_full.append(targ.replace("DAUGHTERN", particle))

        self.edge_conditions = []
        self.edge_transformers = {}
        for i in range(self.Nparticles):
            for j in range(self.Nparticles):
                if i != j:
                    for cond in edge_conditions:
                        self.edge_conditions.append(
                            cond.replace(
                                "DAUGHTERN_DAUGHTERN",
                                f"{self.internal_daughter_particle_names[i]}_{self.internal_daughter_particle_names[j]}",
                            )
                        )
                        self.edge_transformers[
                            cond.replace(
                                "DAUGHTERN_DAUGHTERN",
                                f"{self.internal_daughter_particle_names[i]}_{self.internal_daughter_particle_names[j]}",
                            )
                        ] = self.transformers[cond]

        for idx, cond in enumerate(conditions):
            for i in range(self.Nparticles):
                self.transformers[
                    cond.replace("DAUGHTERN", self.internal_daughter_particle_names[i])
                ] = self.transformers[
                    conditions_raw[idx].replace("DAUGHTERN", "DAUGHTER1")
                ]

        self.query_network = self.query_network_smearing

    def vtx_get_conditional_variables(self):
        mother_conditions = self.rapidsimify_branches(
            self.Model.branch_options["mother_conditions"]
        )
        for idx, cond in enumerate(mother_conditions):
            if cond != "N_daughters":
                self.transformers[cond] = self.transformers[
                    self.Model.branch_options["mother_conditions"][idx]
                ]

        intermediate_conditions = self.rapidsimify_branches(
            self.Model.branch_options["intermediate_conditions"]
        )
        for idx, cond in enumerate(intermediate_conditions):
            if cond != "N_daughters":
                self.transformers[cond] = self.transformers[
                    self.Model.branch_options["intermediate_conditions"][idx].replace(
                        "INTERMEDIATE", "MOTHER"
                    )
                ]

        track_conditions = self.rapidsimify_branches(
            self.Model.branch_options["track_conditions"]
        )
        for idx, cond in enumerate(track_conditions):
            self.transformers[cond.replace("DAUGHTERN", "DAUGHTER1")] = (
                self.transformers[
                    self.Model.branch_options["track_conditions"][idx].replace(
                        "DAUGHTERN", "DAUGHTER1"
                    )
                ]
            )
        return mother_conditions, intermediate_conditions, track_conditions

    def __init__(self, **data):
        super().__init__(**data)

        self.Transformers = self.load_transformers()

        self.internal_daughter_particle_names = list(
            myGlobals.particle_map["fromDAUGHTERS"].keys()
        )
        self.intermediate_particle_name = list(
            myGlobals.particle_map["fromINTERMEDIATES"].keys()
        )

        if "PV_smear" in self.network:
            self.initialise_PV_smearer()
        else:
            self.Model = self.initialise_hetero_model()

            if self.network in [
                "mom_smear",
                "PID_trig",
                "mom_smear_diffusion",
                "PID_trig_diffusion",
                "mom_smear_flowmatching",
                "PID_trig_flowmatching",
            ]:
                self.initialise_unique_node_hetero_model()

            elif "reco_vertex" in self.network:
                self.transformers = self.Transformers

                ##### Organise lists of conditional objects

                self.mother_conditions, intermediate_conditions, track_conditions = (
                    self.vtx_get_conditional_variables()
                )

                self.intermediate_conditions = []
                self.track_conditions = []

                for intermediate in self.intermediate_particle_name:
                    for cond in intermediate_conditions:
                        self.intermediate_conditions.append(
                            cond.replace("INTERMEDIATE", intermediate)
                        )

                for track in self.internal_daughter_particle_names:
                    for cond in track_conditions:
                        self.track_conditions.append(cond.replace("DAUGHTERN", track))

                self.mother_targets = self.Model.branch_options["mother_targets"]
                self.intermediate_targets = []
                self.track_targets = []

                for intermediate in self.intermediate_particle_name:
                    for target in self.Model.branch_options["intermediate_targets"]:
                        self.intermediate_targets.append(
                            target.replace("INTERMEDIATE", intermediate)
                        )

                for track in self.internal_daughter_particle_names:
                    for target in self.Model.branch_options["track_targets"]:
                        self.track_targets.append(target.replace("DAUGHTERN", track))

                ##### Then add transformers

                for track in self.internal_daughter_particle_names:
                    for idx, cond in enumerate(
                        track_conditions + self.Model.branch_options["track_targets"]
                    ):
                        if "DAUGHTERN" in cond:
                            self.transformers[cond.replace("DAUGHTERN", track)] = (
                                self.transformers[
                                    cond.replace("DAUGHTERN", "DAUGHTER1")
                                ]
                            )

                for inter in self.intermediate_particle_name:
                    for cond in self.intermediate_conditions:
                        if "INTERMEDIATE" in cond:
                            pattern = r"INTERMEDIATE\d*"
                            mother_key = re.sub(pattern, "MOTHER", cond)
                            inter_key = re.sub(pattern, inter, cond)
                            self.transformers[inter_key] = self.transformers[mother_key]

                    for targ in self.intermediate_targets:
                        if "INTERMEDIATE" in targ:
                            pattern = r"INTERMEDIATE\d*"
                            mother_key = re.sub(pattern, "MOTHER", targ)
                            inter_key = re.sub(pattern, inter, targ)
                            self.transformers[inter_key] = self.transformers[mother_key]

                # self.query_network = self.query_network_vtx

    def gen_latent(self, chunks, Nparticles):
        # Generate latent noise tensor
        mother_noise = []
        intermediate_noise = []
        track_noise = []
        for i in range(len(chunks)):
            mother_noise.append(np.random.normal(
                0,
                1,
                size=(
                    np.shape(chunks[i])[0],
                    self.Model.gan.mother_latent_dims,
                ),
            ))

            intermediate_noise.append(np.random.normal(
                0,
                1,
                size=(
                    np.shape(chunks[i])[0],
                    self.Model.gan.intermediate_latent_dims,
                ),
            ))

            track_noise.append(np.random.normal(
                0,
                1,
                size=(
                    np.shape(chunks[i])[0],
                    Nparticles,
                    self.Model.gan.track_latent_dims,
                ),
            ))

        return mother_noise, intermediate_noise, track_noise

    def initialise_data_object(self, graph_structure, batch_size):
        empty = torch.tensor([[], []], dtype=torch.long)
        N_nodes = graph_structure.N_nodes

        data = {}
        # self loop
        data["track", "self", "track"] = DataContainer()
        data["track", "self", "track"].edge_index = batchify(
            ("track", "self", "track"),
            torch.tensor([[i, i] for i in range(N_nodes["track"])]).t(),
            batch_size=batch_size,
            N_nodes=N_nodes,
        ).to(myGlobals.device)

        if N_nodes["intermediate"] > 0:
            data["intermediate", "self", "intermediate"] = DataContainer()
            data["intermediate", "self", "intermediate"].edge_index = batchify(
                ("intermediate", "self", "intermediate"),
                torch.tensor([[i, i] for i in range(N_nodes["intermediate"])]).t(),
                batch_size=batch_size,
                N_nodes=N_nodes,
            ).to(myGlobals.device)

        data["mother", "self", "mother"] = DataContainer()
        data["mother", "self", "mother"].edge_index = batchify(
            ("mother", "self", "mother"),
            torch.tensor([[i, i] for i in range(N_nodes["mother"])]).t(),
            batch_size=batch_size,
            N_nodes=N_nodes,
        ).to(myGlobals.device)

        edge_types = [
            ["track", "up", "intermediate"],
            ["track", "up", "mother"],
            ["intermediate", "up", "mother"],
            ["intermediate", "up", "intermediate"],
            ["intermediate", "down", "track"],
            ["mother", "down", "track"],
            ["mother", "down", "intermediate"],
            ["intermediate", "down", "intermediate"],
        ]
        continued = []
        for edge_type in edge_types:
            if edge_type[1] == "up":
                generated_edge_index_tensors = graph_structure.edge_index_tensors_up
            elif edge_type[1] == "down":
                generated_edge_index_tensors = graph_structure.edge_index_tensors_down

            generated_edge_index_tensor = generated_edge_index_tensors[
                (edge_type[0], edge_type[1], edge_type[2])
            ]

            if all(
                isinstance(item, list) and not item
                for item in generated_edge_index_tensor
            ):
                continued.append([edge_type[0], edge_type[1], edge_type[2]])
                continue

            data[edge_type[0], edge_type[1], edge_type[2]] = DataContainer()
            list_of_edges = []
            for edge in generated_edge_index_tensor:
                if edge == []:
                    list_of_edges.append(empty)
                else:
                    list_of_edges.append(
                        batchify(
                            (edge_type[0], edge_type[1], edge_type[2]),
                            torch.tensor(edge).t(),
                            batch_size=batch_size,
                            N_nodes=N_nodes,
                        ).to(myGlobals.device)
                    )
            data[edge_type[0], edge_type[1], edge_type[2]].edge_index = list_of_edges
        for edge_type in continued:
            data[edge_type[0], edge_type[1], edge_type[2]] = DataContainer()
            data[edge_type[0], edge_type[1], edge_type[2]].edge_index = [
                [] for i in range(len(list_of_edges))
            ]

        return data

    def query_network_vtx(
        self,
        graph_structure,
        mother_conditions,
        intermediate_conditions,
        track_conditions,
        Nparticles,
    ):
        with torch.no_grad():
            batch_size = 25000

            mother_chunks = [
                mother_conditions[i : i + batch_size]
                for i in range(0, len(mother_conditions), batch_size)
            ]
            intermediate_chunks = [
                intermediate_conditions[i : i + batch_size]
                for i in range(0, len(intermediate_conditions), batch_size)
            ]
            track_chunks = [
                track_conditions[i : i + batch_size]
                for i in range(0, len(track_conditions), batch_size)
            ]

            # # Handle the last chunk if it's smaller than 1000 rows
            # N_in_final_chunk = int(mother_chunks[-1].shape[0])
            # if N_in_final_chunk < batch_size:
            #     last_chunk = mother_chunks.pop(-1)
            #     shape = np.asarray(np.shape(last_chunk))
            #     shape[0] = batch_size
            #     padded_chunk = np.zeros(tuple(shape))
            #     padded_chunk[: last_chunk.shape[0]] = last_chunk
            #     mother_chunks.append(padded_chunk)

            #     last_chunk = intermediate_chunks.pop(-1)
            #     shape = np.asarray(np.shape(last_chunk))
            #     shape[0] = batch_size
            #     padded_chunk = np.zeros(tuple(shape))
            #     padded_chunk[: last_chunk.shape[0]] = last_chunk
            #     intermediate_chunks.append(padded_chunk)

            #     last_chunk = track_chunks.pop(-1)
            #     shape = np.asarray(np.shape(last_chunk))
            #     shape[0] = batch_size
            #     padded_chunk = np.zeros(tuple(shape))
            #     padded_chunk[: last_chunk.shape[0]] = last_chunk
            #     track_chunks.append(padded_chunk)

            mother_noise, intermediate_noise, track_noise = self.gen_latent(
                mother_chunks, Nparticles
            )

            N_nodes = graph_structure.N_nodes

            if len(mother_noise) > 1:
                data = self.initialise_data_object(graph_structure, batch_size)
            # print(data["mother", "self", "mother"].edge_index)

            self.Model.gan.to(myGlobals.device)

            for i in range(len(mother_noise)):
                print(f"{i}/{len(mother_noise)} chunks processed ...")

                batch_size_i = np.shape(mother_chunks[i])[0]
                if (
                    i + 1 == len(mother_noise) and batch_size_i != batch_size
                ):  # last batch, if different size to others remake edge index tensors
                    data = self.initialise_data_object(graph_structure, batch_size_i)

                track_noise_i = np.reshape(
                    track_noise[i], (-1, self.Model.gan.track_latent_dims)
                )
                track_chunks_i = np.reshape(
                    track_chunks[i], (-1, self.Model.gan.track_conditions_dims)
                )

                if N_nodes["intermediate"] > 0:
                    intermediate_noise_i = intermediate_noise[i]
                    intermediate_chunks_i = intermediate_chunks[i]
                    if N_nodes["intermediate"] > 1:
                        intermediate_noise_i = np.reshape(
                            intermediate_noise_i,
                            (-1, self.Model.gan.intermediate_latent_dims),
                        )
                        intermediate_chunks_i = np.reshape(
                            intermediate_chunks_i,
                            (-1, self.Model.gan.intermediate_conditions_dims),
                        )

                mother_conditions_obj = DataContainer()
                mother_conditions_obj.x = torch.tensor(
                    mother_chunks[i], dtype=torch.float32
                ).to(myGlobals.device)
                track_conditions_obj = DataContainer()
                track_conditions_obj.x = torch.tensor(
                    track_chunks_i, dtype=torch.float32
                ).to(myGlobals.device)
                if N_nodes["intermediate"] > 0:
                    intermediate_conditions_obj = DataContainer()
                    intermediate_conditions_obj.x = torch.tensor(
                        intermediate_chunks_i, dtype=torch.float32
                    ).to(myGlobals.device)

                if "diffusion" in self.network or "flowmatching" in self.network:
                    if "diffusion" in self.network:
                        mother_obj = DataContainer()
                        mother_obj.x = torch.randn(
                            (
                                mother_chunks[i].shape[0],
                                self.Model.gan.mother_targets_dims,
                            ),
                            dtype=torch.float32,
                        ).to(myGlobals.device)

                        track_obj = DataContainer()
                        track_obj.x = torch.randn(
                            (
                                track_chunks_i.shape[0],
                                self.Model.gan.track_targets_dims,
                            ),
                            dtype=torch.float32,
                        ).to(myGlobals.device)

                        if N_nodes["intermediate"] > 0:
                            intermediate_obj = DataContainer()
                            intermediate_obj.x = torch.randn(
                                (
                                    intermediate_chunks_i.shape[0],
                                    self.Model.gan.intermediate_targets_dims,
                                ),
                                dtype=torch.float32,
                            ).to(myGlobals.device)
                    else:
                        mother_obj = DataContainer()
                        mother_obj.x = (
                            torch.randn(
                                (
                                    mother_chunks[i].shape[0],
                                    self.Model.gan.mother_targets_dims,
                                ),
                                dtype=torch.float32,
                            ).to(myGlobals.device)
                            * 0.25
                        )

                        track_obj = DataContainer()
                        track_obj.x = (
                            torch.randn(
                                (
                                    track_chunks_i.shape[0],
                                    self.Model.gan.track_targets_dims,
                                ),
                                dtype=torch.float32,
                            ).to(myGlobals.device)
                            * 0.25
                        )

                        intermediate_obj = DataContainer()
                        intermediate_obj.x = (
                            torch.randn(
                                (
                                    intermediate_chunks_i.shape[0],
                                    self.Model.gan.intermediate_targets_dims,
                                ),
                                dtype=torch.float32,
                            ).to(myGlobals.device)
                            * 0.25
                        )

                    data["mother"] = mother_obj
                    data["track"] = track_obj
                    if N_nodes["intermediate"] > 0:
                        data["intermediate"] = intermediate_obj
                    data["mother_conditions"] = mother_conditions_obj
                    data["track_conditions"] = track_conditions_obj
                    if N_nodes["intermediate"] > 0:
                        data["intermediate_conditions"] = intermediate_conditions_obj
                    self.Model.gan.steps = self.N_diffusion_steps
                    if "flowmatching" in self.network:
                        mother, intermediate, track = self.Model.gan.inference(
                            batch_size_i, data, None, flowmatching=True
                        )
                    else:
                        mother, intermediate, track = self.Model.gan.inference(
                            batch_size_i,
                            data,
                            DDPMScheduler(
                                num_train_timesteps=self.Model.gan.steps,
                                beta_schedule="squaredcos_cap_v2",
                            ),
                            use_encoder=False,
                        )
                    mother = mother.cpu().detach().numpy()
                    if N_nodes["intermediate"] > 0:
                        intermediate = intermediate.cpu().detach().numpy()
                    track = track.cpu().detach().numpy()

                else:
                    mother, intermediate, track = self.Model.gan.generator.forward(
                        torch.tensor(mother_noise[i], dtype=torch.float32).to(
                            myGlobals.device
                        ),
                        torch.tensor(track_noise_i, dtype=torch.float32).to(
                            myGlobals.device
                        ),
                        torch.tensor(intermediate_noise_i, dtype=torch.float32).to(
                            myGlobals.device
                        ),
                        mother_conditions_obj,
                        track_conditions_obj,
                        intermediate_conditions_obj,
                        None,
                        data,  # contains all edge_index information
                    )

                    mother = mother.cpu().detach().numpy()
                    intermediate = intermediate.cpu().detach().numpy()
                    track = track.cpu().detach().numpy()

                if i == 0:
                    if N_nodes["intermediate"] > 0:
                        vertexing_output = np.concatenate(
                            (
                                mother,
                                intermediate.reshape(mother.shape[0], -1),
                                track.reshape(mother.shape[0], -1),
                            ),
                            axis=1,
                        )
                    else:
                        vertexing_output = np.concatenate(
                            (mother, track.reshape(mother.shape[0], -1)),
                            axis=1,
                        )
                else:
                    if N_nodes["intermediate"] > 0:
                        vertexing_output_i = np.concatenate(
                            (
                                mother,
                                intermediate.reshape(mother.shape[0], -1),
                                track.reshape(mother.shape[0], -1),
                            ),
                            axis=1,
                        )
                    else:
                        vertexing_output_i = np.concatenate(
                            (mother, track.reshape(mother.shape[0], -1)),
                            axis=1,
                        )
                    vertexing_output = np.concatenate(
                        (vertexing_output, vertexing_output_i), axis=0
                    )

            # vertexing_output = vertexing_output[
            #     : vertexing_output.shape[0] - (batch_size - N_in_final_chunk)
            # ]

            if N_nodes["intermediate"] > 0:
                out_processed = pd.DataFrame(
                    vertexing_output,
                    columns=self.mother_targets
                    + self.intermediate_targets
                    + self.track_targets,
                )
            else:
                out_processed = pd.DataFrame(
                    vertexing_output,
                    columns=self.mother_targets + self.track_targets,
                )

            out_unprocessed = tfs.untransform_df(out_processed, self.transformers)

            for key in out_unprocessed:
                if "ENDVERTEX_CHI2NDOF" in key:
                    particle = key.replace("_ENDVERTEX_CHI2NDOF", "")
                    if "INTERMEDIATE" in key:
                        particle = myGlobals.particle_map["fromINTERMEDIATES"][
                            particle
                        ]  # daughter_counts dict has user names
                    Ntracks = graph_structure.daughter_counts[particle]
                    NDOF = Ntracks * 2 - 3
                    out_unprocessed[
                        key.replace("ENDVERTEX_CHI2NDOF", "ENDVERTEX_CHI2")
                    ] = out_unprocessed[key] * NDOF

            return out_unprocessed

    def query_network_smearing(self, true_PID_scheme, node_conditions, edge_conditions):
        max_batch_size = 25000

        num_batches = (len(node_conditions) + max_batch_size - 1) // max_batch_size

        # Create lists of batches
        node_conditions_batches = [
            node_conditions[i * max_batch_size : (i + 1) * max_batch_size]
            for i in range(num_batches)
        ]
        edge_conditions_batches = [
            edge_conditions[i * max_batch_size : (i + 1) * max_batch_size]
            for i in range(num_batches)
        ]

        out_unprocessed_batches = []

        for i in range(num_batches):
            node_conditions = node_conditions_batches[i]
            edge_conditions = edge_conditions_batches[i]

            with torch.no_grad():
                dtype = torch.float32
                events_transformed_node = np.asarray(node_conditions).reshape(
                    -1,
                    self.Nparticles,
                    len(self.Model.branch_options["track_conditions"]),
                )
                events_transformed_node = torch.tensor(
                    events_transformed_node, dtype=dtype
                ).to(myGlobals.device)

                N_nodes = {}
                for PID in myGlobals.particle_universe:
                    N_nodes[str(PID)] = 0
                track_PIDs = []
                for particle in true_PID_scheme:
                    if any(
                        myGlobals.particle_map["toDAUGHTERS"][name] in particle
                        for name in self.daughter_particle_names
                    ):
                        PID = abs(true_PID_scheme[particle])
                        N_nodes[str(PID)] += 1
                        track_PIDs.append(PID)

                # possible edges
                ordered_pairs = list(
                    itertools.product(
                        list(map(str, myGlobals.particle_universe)), repeat=2
                    )
                )
                batch_size = events_transformed_node.shape[0]

                data = {}

                for particle_type in myGlobals.particle_universe:
                    data[f"{particle_type}_conditions"] = DataContainer()
                    data[f"{particle_type}_trackidx"] = DataContainer()

                    data[f"{particle_type}_conditions"].x = torch.empty(
                        (0, len(self.Model.branch_options["track_conditions"])),
                        dtype=dtype,
                    ).to(myGlobals.device)

                for track_global_index, track_PID in enumerate(track_PIDs):
                    data[f"{str(track_PID)}_conditions"].x = []
                for track_global_index, track_PID in enumerate(track_PIDs):
                    data[f"{str(track_PID)}_conditions"].x.append(
                        events_transformed_node[:, track_global_index]
                    )
                for particle_type in myGlobals.particle_universe:
                    if particle_type in track_PIDs:
                        if len(data[f"{particle_type}_conditions"].x) == 1:
                            combined = data[f"{particle_type}_conditions"].x[0]
                        else:
                            combined = torch.stack(
                                tuple(data[f"{str(particle_type)}_conditions"].x), dim=1
                            ).to(myGlobals.device)
                            combined = combined.reshape(-1, combined.shape[-1])
                        data[f"{str(particle_type)}_conditions"].x = combined

                particle_types = ["11", "13", "211", "321", "2212"]

                if "diffusion" in self.network or "flowmatching" in self.network:
                    # Add .batch for each particle type
                    for particle_type in particle_types:
                        node_count = data[f"{str(particle_type)}_conditions"].x.shape[0]
                        if node_count > 0:
                            # Repeat batch indices (0 to batch_size-1) for each node in the batch
                            # This assumes the data is laid out in batch-major order
                            # You must know or ensure how data is ordered!
                            counts_per_batch = (
                                node_count // batch_size
                            )  # Assuming equal distribution
                            batch_tensor = torch.arange(
                                batch_size, device=myGlobals.device
                            ).repeat_interleave(counts_per_batch)

                            # If there's a mismatch (edge case), pad or truncate
                            if batch_tensor.shape[0] < node_count:
                                extra = torch.full(
                                    (node_count - batch_tensor.shape[0],),
                                    batch_size - 1,
                                    device=myGlobals.device,
                                )
                                batch_tensor = torch.cat([batch_tensor, extra])
                            elif batch_tensor.shape[0] > node_count:
                                batch_tensor = batch_tensor[:node_count]

                            data[particle_type] = (
                                DataContainer()
                                if particle_type not in data
                                else data[particle_type]
                            )
                            data[particle_type].batch = batch_tensor
                        else:
                            data[particle_type] = DataContainer()

                for pair in ordered_pairs:
                    data[pair[0], "to", pair[1]] = DataContainer()

                for pair in ordered_pairs:
                    data[pair[0], "to", pair[1]].edge_index = torch.empty(
                        (2, 0), dtype=torch.long
                    ).to(myGlobals.device)
                    data[pair[0], "to", pair[1]].edge_attr = []

                ##### empty batch object created...

                edges_by_index = list(
                    itertools.product(range(len(track_PIDs)), repeat=2)
                )

                index_by_appearence = []
                index_counters = {11: 0, 13: 0, 211: 0, 321: 0, 2212: 0}
                for item in track_PIDs:
                    index_by_appearence.append(index_counters[item])
                    index_counters[item] += 1

                for edge_by_index in edges_by_index:
                    origin = str(track_PIDs[edge_by_index[0]])
                    destination = str(track_PIDs[edge_by_index[1]])
                    origin_idx = index_by_appearence[edge_by_index[0]]
                    destination_idx = index_by_appearence[edge_by_index[1]]

                    if origin != destination or origin_idx != destination_idx:
                        origin_name = self.internal_daughter_particle_names[
                            edge_by_index[0]
                        ]
                        destination_name = self.internal_daughter_particle_names[
                            edge_by_index[1]
                        ]

                        edge_conditions_i = [
                            cond
                            for cond in self.edge_conditions
                            if f"{origin_name}_{destination_name}" in cond
                        ]

                        edge_attr_i = torch.tensor(
                            np.asarray(edge_conditions[edge_conditions_i]), dtype=dtype
                        ).to(myGlobals.device)
                        data[origin, "to", destination].edge_attr.append(edge_attr_i)

                        edge_i = (
                            torch.tensor(
                                [[origin_idx, destination_idx]], dtype=torch.long
                            )
                            .t()
                            .to(myGlobals.device)
                        )
                        data[origin, "to", destination].edge_index = torch.cat(
                            (data[origin, "to", destination].edge_index, edge_i), dim=1
                        )

                for pair in ordered_pairs:
                    if len(data[pair[0], "to", pair[1]].edge_attr) == 0:
                        data[pair[0], "to", pair[1]].edge_attr = torch.empty(
                            (0, len(self.Model.branch_options["edge_conditions"])),
                            dtype=dtype,
                        ).to(myGlobals.device)
                    else:
                        combined = (
                            torch.stack(
                                tuple(data[pair[0], "to", pair[1]].edge_attr), dim=1
                            )
                            .reshape(-1, 1)
                            .to(myGlobals.device)
                        )
                        data[pair[0], "to", pair[1]].edge_attr = combined

                for pair in ordered_pairs:
                    if data[pair[0], "to", pair[1]].edge_index.shape[1] > 0:
                        data[pair[0], "to", pair[1]].edge_index = batchify(
                            (pair[0], "to", pair[1]),
                            data[pair[0], "to", pair[1]].edge_index,
                            batch_size=batch_size,
                            N_nodes=N_nodes,
                        )

                track_latent = {}
                for particle_type in myGlobals.particle_universe:
                    track_latent[str(particle_type)] = torch.randn(
                        (
                            data[f"{str(particle_type)}_conditions"].x.shape[0],
                            self.Model.gan.track_latent_dims,
                        ),
                        dtype=dtype,
                    ).to(myGlobals.device)

                self.Model.gan.to(myGlobals.device)

                if "diffusion" in self.network:
                    for particle_type in particle_types:
                        # data[particle_type] = DataContainer()
                        data[particle_type].x = torch.randn(
                            (
                                data[f"{particle_type}_conditions"].x.shape[0],
                                self.Model.gan.track_targets_dims,
                            ),
                            dtype=dtype,
                        ).to(myGlobals.device)

                    self.Model.gan.diffuser.to(myGlobals.device)

                    self.Model.gan.steps = self.N_diffusion_steps
                    if "PID" in self.network:
                        out = self.Model.gan.inference(
                            batch_size,
                            data,
                            DDPMScheduler(
                                num_train_timesteps=self.Model.gan.steps,
                                beta_schedule="squaredcos_cap_v2",
                            ),
                        )
                    else:
                        out = self.Model.gan.inference(
                            batch_size,
                            data,
                            DDPMScheduler(
                                num_train_timesteps=self.Model.gan.steps,
                                clip_sample=True,
                                clip_sample_range=5.5,
                                beta_schedule="squaredcos_cap_v2",
                            ),
                        )

                    # out = out.cpu().detach().numpy()
                    # out = {k: v.cpu().detach().numpy() for k, v in out.items()}
                elif "flowmatching" in self.network:
                    for particle_type in particle_types:
                        data[particle_type].x = torch.randn(
                            (
                                data[f"{particle_type}_conditions"].x.shape[0],
                                self.Model.gan.track_targets_dims,
                            ),
                            dtype=dtype,
                        ).to(myGlobals.device)
                    self.Model.gan.diffuser.to(myGlobals.device)
                    out = self.Model.gan.inference(
                        batch_size, data, None, flowmatching=True
                    )

                else:
                    out = self.Model.gan.generator(
                        track_latent,
                        batch=data,
                    )

                for i in out:
                    i_out = out[i].cpu().detach().numpy()
                    for idx in range(len(self.targets)):
                        if myGlobals.personalised_track_node_types:
                            i_out[:, idx] = self.Transformers[
                                f"{self.targets[idx].replace('DAUGHTERN', 'DAUGHTER1')}_{i}"
                            ].unprocess(i_out[:, idx])
                        else:
                            i_out[:, idx] = self.Transformers[
                                f"{self.targets[idx].replace('DAUGHTERN', 'DAUGHTER1')}"
                            ].unprocess(i_out[:, idx])
                    out[i] = i_out

                for i in out:
                    if N_nodes[i] > 0:
                        out[i] = out[i].reshape(-1, N_nodes[i], len(self.targets))

                N_nodes_out = {}
                for PID in myGlobals.particle_universe:
                    N_nodes_out[str(PID)] = 0

                for idx, item in enumerate(track_PIDs):
                    if idx == 0:
                        out_ordered = np.expand_dims(out[str(item)][:, 0], 1)
                    else:
                        out_ordered = np.concatenate(
                            (
                                out_ordered,
                                np.expand_dims(
                                    out[str(item)][:, N_nodes_out[str(item)]], 1
                                ),
                            ),
                            axis=1,
                        )
                    N_nodes_out[str(item)] += 1

                out_ordered = out_ordered.reshape(
                    -1, len(self.targets) * len(track_PIDs)
                )

                out_unprocessed = pd.DataFrame(out_ordered, columns=self.targets_full)

                out_unprocessed_batches.append(out_unprocessed)

        out_unprocessed = pd.concat(out_unprocessed_batches, ignore_index=True)

        return out_unprocessed

    def query_network_vanilla(self, conditions):
        if "diffusion" in self.network:
            with torch.no_grad():
                self.Model.gan.to(myGlobals.device)

                conditions = torch.tensor(conditions, dtype=torch.float32).to(
                    myGlobals.device
                )

                reconstructed = self.Model.gan.inference(
                    conditions.shape[0], conditions
                )

                reconstructed = reconstructed.detach().cpu().numpy()
                reconstructed = pd.DataFrame(reconstructed, columns=self.targets)
                output = tfs.untransform_df(reconstructed, self.transformers)

            return output

        else:
            with torch.no_grad():
                self.Model.gan.to(myGlobals.device)

                conditions = torch.tensor(conditions, dtype=torch.float32).to(
                    myGlobals.device
                )
                z = torch.randn((conditions.shape[0], self.Model.gan.latent_dims)).to(
                    myGlobals.device
                )

                reconstructed = self.Model.gan.generator(z, conditions)

                reconstructed = reconstructed.detach().cpu().numpy()
                reconstructed = pd.DataFrame(reconstructed, columns=self.targets)
                output = tfs.untransform_df(reconstructed, self.transformers)

            return output



import lhcb_rex.data.graph_maker as graph_maker
import argparse
import json
import lhcb_rex.settings.globals as myGlobals


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


# --processID 0 --config_file "./davinci/configs/configs_4_body/config_dict_B0_AAA+int(DDD-int(->BBB-CCC+)).pickle" --file_name "/dice/users/am13743/fast_vertex_quality/GENERALISED/N4_topIdx4_head_more_vars.root" --splits 5


def parse_args():
    parser = argparse.ArgumentParser(
        description="Load settings from command line arguments."
    )
    parser.add_argument(
        "--file_name", type=str, required=True, help="Path to the input ROOT file"
    )
    parser.add_argument(
        "--config_file", type=str, required=True, help="Path to the input ROOT file"
    )
    parser.add_argument("--root", type=str, required=True, help="Root directory path")
    parser.add_argument("--splits", type=int, required=True)
    parser.add_argument("--save_string", type=str, required=True)
    parser.add_argument("--extra_tag", type=str, required=True)
    parser.add_argument("--use_weights", type=str2bool, required=True)
    parser.add_argument("--smearingnet", type=str2bool, required=True)
    parser.add_argument("--PIDnet", type=str2bool, required=True)
    parser.add_argument("--processID", type=int, required=True)
    parser.add_argument("--require_validation_variables", type=str2bool, required=True)
    return parser.parse_args()


args = parse_args()

output_path = args.root + f"_{args.extra_tag}/"

if args.smearingnet:
    mother_targets = []
    intermediate_targets = []
    track_targets = myGlobals.smearing_track_targets
    mother_conditions = []
    intermediate_conditions = []
    track_conditions = myGlobals.smearing_track_conditions
    edge_conditions = myGlobals.smearing_edge_conditions
elif args.PIDnet:
    mother_targets = []
    intermediate_targets = []
    track_targets = myGlobals.PID_track_targets
    mother_conditions = []
    intermediate_conditions = []
    track_conditions = myGlobals.PID_track_conditions
    edge_conditions = myGlobals.PID_edge_conditions
else:
    mother_targets = myGlobals.mother_targets
    intermediate_targets = myGlobals.intermediate_targets
    track_targets = myGlobals.track_targets
    mother_conditions = myGlobals.mother_conditions
    intermediate_conditions = myGlobals.intermediate_conditions
    track_conditions = myGlobals.track_conditions
    edge_conditions = []


graph_maker.GENERALISED_HeteroGraphDataset(
    path=args.file_name,
    root=output_path,
    config=args.config_file,
    splits=args.splits,
    use_weights=args.use_weights,
    smearingnet=args.smearingnet,
    PIDnet=args.PIDnet,
    processID=args.processID,
    require_validation_variables=args.require_validation_variables,
    force_reload=True,
    mother_targets=mother_targets,
    intermediate_targets=intermediate_targets,
    track_targets=track_targets,
    mother_conditions=mother_conditions,
    intermediate_conditions=intermediate_conditions,
    track_conditions=track_conditions,
    edge_conditions=edge_conditions,
)

# quit()


# settings = {}
# settings[args.save_string] = {
#     "file_name": args.file_name,
#     "particles_involved": args.particles_involved,
#     "intermediates": json.loads(args.intermediates),
#     "mother_N": args.mother_N,
#     "intermediate_N": json.loads(args.intermediate_N),
#     "root": args.root,
# }

# print(json.dumps(settings, indent=4))

# max_N = args.max_N
# splits = args.splits
# extra_tag = args.extra_tag
# use_weights = args.use_weights
# full_reco = args.full_reco
# smearingnet = args.smearingnet
# require_validation_variables = args.require_validation_variables
# PIDnet = args.PIDnet

# if smearingnet and PIDnet:
#     print("smearingnet and PIDnet, pick one")
#     raise Exception

# print(max_N)
# print(splits)
# print(extra_tag)
# print(use_weights)
# print(full_reco)


# processID = args.processID
# print("processID", processID)

# # edge_conditions = [
# #     "edge_angle_DAUGHTERN_DAUGHTERN",
# #     "edge_angle_DAUGHTERN_DAUGHTERN_TRUE",
# # ]

# for setting in settings:
#     print(f"\n\n\n {settings[setting]}")

#     file_name = settings[setting]["file_name"]
#     particles_involved = settings[setting]["particles_involved"]
#     intermediates = settings[setting]["intermediates"]
#     mother_N = settings[setting]["mother_N"]
#     intermediate_N = settings[setting]["intermediate_N"]
#     root = settings[setting]["root"]

#     print(f"\n {full_reco}")

#     if full_reco:
#         reco_tag = "fullreco"
#     else:
#         reco_tag = "partreco"

#     splits_in = splits

#     output_path = root + f"_{reco_tag}{extra_tag}/"
#     if smearingnet:
#         output_path = root + f"_{reco_tag}{extra_tag}smearingnet/"

#     if smearingnet:
#         mother_targets = []
#         intermediate_targets = []
#         track_targets = myGlobals.smearing_track_targets
#         mother_conditions = []
#         intermediate_conditions = []
#         track_conditions = myGlobals.smearing_track_conditions
#         edge_conditions = myGlobals.smearing_edge_conditions
#     elif PIDnet:
#         mother_targets = []
#         intermediate_targets = []
#         track_targets = myGlobals.PID_track_targets
#         mother_conditions = []
#         intermediate_conditions = []
#         track_conditions = myGlobals.PID_track_conditions
#         edge_conditions = myGlobals.PID_edge_conditions
#     else:
#         mother_targets = myGlobals.mother_targets
#         intermediate_targets = myGlobals.intermediate_targets
#         track_targets = myGlobals.track_targets
#         mother_conditions = myGlobals.mother_conditions
#         intermediate_conditions = myGlobals.intermediate_conditions
#         track_conditions = myGlobals.track_conditions
#         edge_conditions = []

#     graph_maker.HeteroGraphDataset(
#         root=output_path,
#         particles_involved=particles_involved,
#         intermediates=intermediates,
#         fully_reco=full_reco,
#         max_N_samples_per=max_N,
#         force_reload=True,
#         path=file_name,
#         mother_targets=mother_targets,
#         intermediate_targets=intermediate_targets,
#         track_targets=track_targets,
#         mother_conditions=mother_conditions,
#         intermediate_conditions=intermediate_conditions,
#         track_conditions=track_conditions,
#         edge_conditions=edge_conditions,
#         use_weights=use_weights,
#         splits=splits_in,
#         mother_N=mother_N,
#         intermediate_N=intermediate_N,
#         processID=processID,
#         smearingnet=smearingnet,
#         PIDnet=PIDnet,
#         require_validation_variables=require_validation_variables,
#     )



from huggingface_hub import snapshot_download
from appdirs import user_cache_dir
import os
import glob
import lhcb_rex.tools.display as display
from pathlib import Path

class HuggingFacePath:
    def __init__(self, path):
        self.path = path

def get_model_paths():

    with (
        display.status_execution(
            status_message="[bold green]Downloading/checking for network weights...",
            complete_message="[bold green]Complete.:white_check_mark:",
        )
    ):
        
        if "CI_PROJECT_DIR" in os.environ:
            base_cache = Path(os.environ["CI_PROJECT_DIR"]) / ".cache" / "lhcb_rex"
        else:
            base_cache = user_cache_dir("lhcb_rex")
        cache_dir = os.path.join(base_cache, "weights")
        model_path = snapshot_download(
            repo_id="alexmarshallbristol/Rex",
            cache_dir=cache_dir,
            local_dir_use_symlinks=True,
        )
        weight_files = glob.glob(f'{model_path}/*.pkl')

        paths = {}
        for weight_file in weight_files:
            paths[weight_file.split('/')[-1].replace('.pkl','')] = HuggingFacePath(weight_file)

    display.info_print(f'Weights stored at: {Path(model_path).parent.parent}')

    return paths
